#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import fsspec
import os

import click
import yaml

from ads.common.auth import AuthType
from ads.common import auth as authutil
from ads.opctl.cmds import activate as activate_cmd
from ads.opctl.cmds import cancel as cancel_cmd
from ads.opctl.cmds import deactivate as deactivate_cmd
from ads.opctl.cmds import configure as configure_cmd
from ads.opctl.cmds import delete as delete_cmd
from ads.opctl.cmds import init_vscode as init_vscode_cmd
from ads.opctl.cmds import run as run_cmd
from ads.opctl.cmds import run_diagnostics as run_diagnostics_cmd
from ads.opctl.cmds import watch as watch_cmd
from ads.opctl.cmds import init_operator as init_operator_cmd
from ads.opctl.utils import build_image as build_image_cmd
from ads.opctl.utils import publish_image as publish_image_cmd
from ads.opctl.utils import suppress_traceback
from ads.opctl.config.merger import ConfigMerger
from ads.opctl.constants import BACKEND_NAME

import ads.opctl.conda.cli
import ads.opctl.spark.cli
import ads.opctl.distributed.cli


@click.group("opctl")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def configure(debug):
    suppress_traceback(debug)(configure_cmd)()


@commands.command()
@click.argument("image-type", type=click.Choice(["job-local", "ads-ops-base"]))
@click.help_option("--help", "-h")
@click.option(
    "--gpu",
    "-g",
    help="whether to build a gpu version of image",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--source-folder",
    "-s",
    help="when building custom operator image, source folder of the custom operator",
    default=None,
    required=False,
)
@click.option(
    "--image",
    "-i",
    help="image name, used when building custom image",
    default=None,
    required=False,
)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def build_image(image_type, gpu, source_folder, image, debug):
    suppress_traceback(debug)(build_image_cmd)(image_type, gpu, source_folder, image)


@commands.command()
@click.argument("image")
@click.option(
    "--registry", "-r", help="registry to push to", required=False, default=None
)
@click.option(
    "--ads-config",
    help="folder that saves ads opctl config",
    required=False,
    default=None,
)
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def publish_image(**kwargs):
    debug = kwargs.pop("debug")
    if kwargs.get("registry", None):
        registry = kwargs["registry"]
    else:
        p = ConfigMerger({}).process(**kwargs)
        registry = p.config.get("infrastructure", {}).get("docker_registry", None)
    suppress_traceback(debug)(publish_image_cmd)(kwargs["image"], registry)


@commands.command()
@click.option(
    "--source-folder",
    "-s",
    help="source folder for scripts",
    required=False,
    default=None,
)
@click.option(
    "--image",
    "-i",
    help="docker image name",
    required=False,
    default=None,
)
@click.option(
    "--oci-config",
    help="oci config file",
    required=False,
    default=None,
)
@click.option(
    "--conda-pack-folder",
    help="folder where conda packs are saved",
    required=False,
    default=None,
)
@click.option(
    "--ads-config",
    help="folder that saves ads opctl config",
    required=False,
    default=None,
)
@click.option(
    "--gpu",
    "-g",
    help="whether to run against GPU image",
    is_flag=True,
    default=False,
    required=False,
)
@click.option(
    "--env-var",
    help="environment variables to pass in",
    multiple=True,
    required=False,
    default=None,
)
@click.option(
    "--volume",
    "-v",
    help="volumes to mount to image",
    multiple=True,
    required=False,
    default=None,
)
@click.help_option("--help", "-h")
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
def init_vscode(**kwargs):
    """
    Generates devcontainer.json with docker details to setup a development environment for OCI Data Science Jobs
    """
    # kwargs["use_conda"] = True
    suppress_traceback(kwargs["debug"])(init_vscode_cmd)(**kwargs)


_options = [
    click.option(
        "--file", "-f", help="path to resource YAML file", required=False, default=None
    ),
    click.option(
        "--operator-name", "-n", help="operator name", required=False, default=None
    ),
    click.option(
        "--backend",
        "-b",
        type=click.Choice([backend.value for backend in BACKEND_NAME]),
        help="backend to run the operator",
        required=False,
        default=None,
    ),
    click.option(
        "--oci-config",
        help="oci config file",
        required=False,
        default=None,
    ),
    click.option(
        "--oci-profile", help="oci config profile", required=False, default=None
    ),
    click.option(
        "--conf-file",
        help="path to conf file",
        required=False,
        default=None,
    ),
    click.option(
        "--conf-profile",
        help="conf profile",
        required=False,
        default=None,
    ),
    click.option(
        "--conda-pack-folder",
        help="folder where conda packs are saved",
        required=False,
        default=None,
    ),
    click.option(
        "--ads-config",
        help="folder that saves ads opctl config",
        required=False,
        default=None,
    ),
    click.help_option("--help", "-h"),
    click.option(
        "--job-name", help="display name of job", required=False, default=None
    ),
    click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False),
    click.option(
        "--auth",
        "-a",
        help="authentication method",
        type=click.Choice(["api_key", "resource_principal"]),
        default=None,
    ),
    click.option(
        "--wait-for-completion",
        help="either to wait for process to complete or not",
        is_flag=True,
        required=False,
    ),
    click.option(
        "--max-wait-time",
        help="maximum wait time in seconds for progress to complete",
        type=int,
        required=False,
        default=1200,
    ),
    click.option(
        "--poll-interval",
        help="poll interval in seconds",
        type=int,
        required=False,
        default=10,
    ),
    click.option(
        "--log-type", help="the type of logging.", required=False, default=None
    ),
    click.option(
        "--log-filter",
        help="expression for filtering the logs.",
        required=False,
        default=None,
    ),
    click.option(
        "--interval",
        help="log interval in seconds",
        type=int,
        required=False,
        default=3,
    ),
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@commands.command()
@add_options(_options)
@click.option("--image", "-i", help="image name", required=False, default=None)
@click.option("--conda-slug", help="slug name", required=False, default=None)
@click.option(
    "--use-conda",
    help="whether to use conda pack to run an operator",
    is_flag=True,
    required=False,
    default=None,
)
@click.option(
    "--source-folder",
    "-s",
    help="source scripts folder that will be mounted during a local run",
    required=False,
    default=None,
)
@click.option(
    "--env-var",
    help="environment variables to pass in",
    multiple=True,
    required=False,
    default=None,
)
@click.option(
    "--entrypoint",
    "-e",
    help="entrypoint, a script in case of running with conda pack or entrypoint in case of running with an image",
    required=False,
    default=None,
)
@click.option(
    "--command",
    "-c",
    help="docker command, used when running with an image",
    required=False,
    default=None,
)
@click.option(
    "--cmd-args",
    help="command line args, used when running with conda pack",
    required=False,
    default=None,
)
@click.option(
    "--gpu",
    "-g",
    help="whether to run against GPU image",
    is_flag=True,
    default=None,
    required=False,
)
@click.option(
    "--volume",
    "-v",
    help="volumes to mount to image, used only for local testing at the moment",
    multiple=True,
    required=False,
    default=None,
)
@click.option(
    "--overwrite",
    "-o",
    help="overwrite object storage files when uploading local files during dataflow run",
    is_flag=True,
    default=False,
)
@click.option("--archive", help="path to archive zip for dataflow run", default=None)
@click.option("--exclude-tag", multiple=True, default=None)
@click.option(
    "--dry-run",
    "-r",
    default=False,
    is_flag=True,
    help="During dry run, the actual operation is not performed, only the steps are enumerated.",
)
@click.option(
    "--ocid",
    required=False,
    default=None,
    help="create an OCI Data Science service run from existing ocid.",
)
@click.option(
    "--nobuild",
    "-nobuild",
    default=False,
    is_flag=True,
    help="skip re-building the docker image",
)
@click.option(
    "--auto_increment",
    default=False,
    is_flag=True,
    help="Increments tag of the image while rebuilding",
)
@click.option(
    "--nopush",
    "-nopush",
    default=False,
    is_flag=True,
    help="Image is not pushed to OCIR",
)
@click.option("--tag", "-t", help="tag of image", required=False, default=None)
@click.option(
    "--registry", "-reg", help="registry to push to", required=False, default=None
)
@click.option(
    "--dockerfile",
    "-df",
    help="relative path to Dockerfile",
    required=False,
    default=None,
)
def run(file, **kwargs):
    """
    Runs the workload on the targeted backend. When run `distributed` yaml spec, the backend is always OCI Data Science
    Jobs
    """
    debug = kwargs["debug"]
    if file:
        if os.path.exists(file):
            auth = {}
            if kwargs["auth"]:
                auth = authutil.create_signer(kwargs["auth"])
            else:
                auth = authutil.default_signer()

            with fsspec.open(file, "r", **auth) as f:
                config = suppress_traceback(debug)(yaml.safe_load)(f.read())
        else:
            raise FileNotFoundError(f"{file} is not found")
    else:
        # If no yaml is provided, we assume there's cmdline args to define a job.
        config = {"kind": "job"}
    suppress_traceback(debug)(run_cmd)(config, **kwargs)


@commands.command()
@add_options(_options)
@click.option(
    "--dry-run",
    "-r",
    default=False,
    is_flag=True,
    help="During dry run, the actual operation is not performed, only the steps are enumerated.",
)
@click.option(
    "--output", help="filename to save the report", default="diagnostic_report.html"
)
def check(file, **kwargs):
    """
    Run diagnostics to check if your setup meets all the requirements to run the job
    """
    debug = kwargs["debug"]
    if file:
        if os.path.exists(file):
            with open(file, "r") as f:
                config = suppress_traceback(debug)(yaml.safe_load)(f.read())
        else:
            raise FileNotFoundError(f"{file} is not found")
    else:
        config = {}
    suppress_traceback(debug)(run_diagnostics_cmd)(config, **kwargs)


@commands.command()
@click.argument("operator_slug", nargs=1)
@click.option(
    "--folder_path",
    "-fp",
    help="the name of the folder wherein to put the operator code",
    multiple=True,
    required=False,
    default=None,
)
@add_options(_options)
def init_operator(**kwargs):
    suppress_traceback(kwargs["debug"])(init_operator_cmd)(**kwargs)


@commands.command()
@click.argument("ocid", nargs=1)
@add_options(_options)
def delete(**kwargs):
    suppress_traceback(kwargs["debug"])(delete_cmd)(**kwargs)


@commands.command()
@click.argument("ocid", nargs=1)
@add_options(_options)
def cancel(**kwargs):
    suppress_traceback(kwargs["debug"])(cancel_cmd)(**kwargs)


@commands.command()
@click.argument("ocid", nargs=1)
@add_options(_options)
def watch(**kwargs):
    """
    ``tail`` logs form a job run, dataflow run or pipeline run.
    Connects to the logging service that was configured with the JobRun, Application Run or Pipeline Run and streams the logs.
    """
    suppress_traceback(kwargs["debug"])(watch_cmd)(**kwargs)


@commands.command()
@click.argument("ocid", nargs=1)
@add_options(_options)
def activate(**kwargs):
    """
    Activates a data science service.
    """
    suppress_traceback(kwargs["debug"])(activate_cmd)(**kwargs)


@commands.command()
@click.argument("ocid", nargs=1)
@add_options(_options)
def deactivate(**kwargs):
    """
    Deactivates a data science service.
    """
    suppress_traceback(kwargs["debug"])(deactivate_cmd)(**kwargs)


commands.add_command(ads.opctl.conda.cli.commands)
commands.add_command(ads.opctl.spark.cli.commands)
commands.add_command(ads.opctl.distributed.cli.commands)

# @commands.command()
# @click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
# def list(debug):
#     print(json.dumps(suppress_traceback(debug)(_list_ads_operators)(), indent=2))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import re
import string
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import ByteString, Dict, List, Optional, Union

import fsspec
from oci.data_science.models.pipeline_step_run import PipelineStepRun


class PipelineVisualizerError(Exception):
    pass


class StepStatus:
    WAITING = "WAITING"
    ACCEPTED = "ACCEPTED"
    IN_PROGRESS = "IN_PROGRESS"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"
    CANCELING = "CANCELING"
    CANCELED = "CANCELED"
    DELETED = "DELETED"
    SKIPPED = "SKIPPED"


class GraphOrientation:
    TOP_BOTTOM = "TB"
    LEFT_RIGHT = "LR"


WAIT_STATUS = {
    StepStatus.WAITING,
    StepStatus.IN_PROGRESS,
    StepStatus.CANCELING,
}

COMPLETE_STATUS = {
    StepStatus.SUCCEEDED,
    StepStatus.FAILED,
    StepStatus.CANCELED,
    StepStatus.DELETED,
    StepStatus.SKIPPED,
}

FAIL_STATUS = {
    StepStatus.FAILED,
    StepStatus.CANCELED,
    StepStatus.DELETED,
    StepStatus.SKIPPED,
}

STATUS_TEXT_MAP = {
    StepStatus.WAITING: "Waiting",
    StepStatus.ACCEPTED: "Accepted",
    StepStatus.IN_PROGRESS: "In Progress",
    StepStatus.FAILED: "Failed",
    StepStatus.SUCCEEDED: "Succeeded",
    StepStatus.CANCELING: "Canceling",
    StepStatus.CANCELED: "Canceled",
    StepStatus.DELETED: "Deleted",
    StepStatus.SKIPPED: "Skipped",
}

STATUS_COLOR_MAP = {
    StepStatus.WAITING: "#747E7E",
    StepStatus.ACCEPTED: "#F26B1D",
    StepStatus.IN_PROGRESS: "#3E975D",
    StepStatus.FAILED: "#9146C2",
    StepStatus.SUCCEEDED: "#2C6CBF",
    StepStatus.CANCELING: "#D92211",
    StepStatus.CANCELED: "#D92211",
    StepStatus.DELETED: "#D92211",
    StepStatus.SKIPPED: "#D92211",
}


class StepKind:
    ML_JOB = "ML_JOB"
    CUSTOM_SCRIPT = "CUSTOM_SCRIPT"
    PIPELINE = "pipeline"


STEP_KIND_MAP = {
    StepKind.ML_JOB: "Datascience Job",
    StepKind.CUSTOM_SCRIPT: "Custom Script",
    StepKind.PIPELINE: "Pipeline",
}


def _to_utc_native(dt: datetime) -> str:
    """Converts offset-aware datetime to offset-native UTC time.

    Parameters
    ----------
    dt : datetime.datetime
        A datetime object to be converted.
        If dt is offset native, it will be returned as is.

    Returns
    -------
    datetime.datetime
        Offset-native datetime
    """
    if not dt:
        return None

    if dt.tzinfo:
        return dt.astimezone(timezone.utc).replace(tzinfo=None)
    return dt


def _strfdelta(seconds: int, format: str = "%H:%M:%S") -> str:
    """Converts a datetime.timedelta object to a custom-formatted string.

    Parameters
    ----------
    seconds: int
        Seconds to be converted to a formatted string.
    format: (str, optional). Defaults to `%H:%M:%S`.
        The format argument allows custom formatting to be specified.
    """
    if not seconds:
        return ""
    return datetime.utcfromtimestamp(seconds).strftime(format)


def _replace_special_chars(value: str, repl: Optional[str] = "_") -> str:
    """Returns a string obtained by replacing the all special characters in the string,
    including spaces, by the replacement `repl`.

    Parameters
    ----------
    value: str
        The string that needs to be processed.
    repl: (str, optional). Defaults to `_`.
        The replacement for the special characters.

    Returns
    -------
    str
        The new string.
    """
    return re.sub(f"[{re.escape(string.punctuation)} ]", repl, value)


@dataclass
class RendererItemStatus:
    """Class represents the state of the renderer item."""

    name: str
    kind: str = ""
    time_started: datetime = None
    time_finished: datetime = None
    lifecycle_state: str = ""
    lifecycle_details: str = ""
    _key: str = ""

    def __post_init__(self):
        self.time_started = _to_utc_native(self.time_started)
        self.time_finished = _to_utc_native(self.time_finished)
        self._key = f"{_replace_special_chars(self.name)}_{self.kind}".lower()

    @property
    def key(self) -> str:
        """Key of the item.

        Returns
        -------
        str
            The key of the item.
        """
        return self._key

    @property
    def duration(self) -> int:
        """
        Calculates duration in seconds between `time_started` and `time_finished`.

        Returns
        -------
        int
            The duration in seconds between `time_started` and `time_finished`.
        """
        if not self.time_started:
            return 0
        if not self.time_finished:
            return (datetime.utcnow() - self.time_started).seconds
        return (self.time_finished - self.time_started).seconds

    @classmethod
    def from_pipeline_run(cls, pipeline_run: "PipelineRun") -> "RendererItemStatus":
        """Creates class instance from the PipelineRun object.

        Parameters
        ----------
        pipeline_run: PipelineRun
            The PipelineRun object.

        Returns
        -------
        RendererItemStatus
            Instance of RendererItemStatus.

        """
        return cls(
            name=pipeline_run.pipeline.name,
            kind=pipeline_run.pipeline.kind,
            time_started=pipeline_run.time_accepted or pipeline_run.time_started,
            time_finished=pipeline_run.time_finished,
            lifecycle_state=pipeline_run.lifecycle_state,
            lifecycle_details=pipeline_run.lifecycle_details,
        )

    @classmethod
    def from_pipeline_step_run(
        cls, pipeline_step_run: PipelineStepRun
    ) -> "RendererItemStatus":
        """Creates class instance from the PipelineStepRun object.

        Parameters
        ----------
        pipeline_run: PipelineStepRun
            The PipelineStepRun object.

        Returns
        -------
        RendererItemStatus
            Instance of RendererItemStatus.

        """
        return cls(
            name=pipeline_step_run.step_name,
            kind=pipeline_step_run.step_type,
            time_started=pipeline_step_run.time_started,
            time_finished=pipeline_step_run.time_finished,
            lifecycle_state=pipeline_step_run.lifecycle_state,
            lifecycle_details=pipeline_step_run.lifecycle_details,
        )

    @staticmethod
    def format_datetime(value: datetime, format="%Y-%m-%d %H:%M:%S") -> str:
        """Converts datetime object into a given format in string

        Parameters
        ----------
        dt: datetime.datetime
            Datetime object to be formated.

        Returns
        -------
        str:
            A timestamp in a string format.
        """
        if not value:
            return None

        return value.strftime(format)

    def __hash__(self):
        return hash(self._key)

    def __eq__(self, other):
        return hash(self) == hash(other)


@dataclass
class RendererItem:
    name: str
    kind: str = ""
    spec: Union["Pipeline", "PipelineStep"] = None
    _key: str = ""

    def __post_init__(self):
        self._key = f"{_replace_special_chars(self.name)}_{self.kind}".lower()

    @property
    def key(self) -> str:
        """Key of the item.

        Returns
        -------
        str
            The key of the item.
        """
        return self._key

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return hash(self) == hash(other)


class PipelineRenderer(ABC):
    """The base class responsible for the vizualizing a pipleine."""

    @abstractmethod
    def render(
        self,
        steps: List[RendererItem],
        deps: Dict[str, List[RendererItem]],
        step_status: Dict[str, List[RendererItemStatus]] = None,
        **kwargs,
    ):
        """Renders pipeline run."""
        pass

    def _write_to_file(
        self, content: Union[str, ByteString], mode: str, uri: str, **kwargs
    ) -> None:
        """Writes string into location specified by uri.

        Parameters
        ----------
        content: Union[str, ByteString]
            The content to be saved.
        mode: str
            The file write mode.
        uri: (string)
            URI location to save content.

        kwargs
        ------
        keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
            For other storage connections consider e.g. host, port, username, password, etc.

        Returns
        -------
        None
            Nothing.
        """
        with fsspec.open(uri, f"w{mode}", **kwargs) as f:
            f.write(content)

    def save_to(self, *args, **kwargs) -> str:
        """Saves the pipeline visualization to the provided format."""
        raise NotImplementedError("`.save_to()` is not implemented.")


class PipelineVisualizer:
    """PipelineVisualizer class to visualize pipeline in text or graph.

    Attributes
    ----------
    pipeline: Pipeline
        Pipeline instance.
    pipeline_run: PipelineRun
        PipelineRun instance.
    steps: List[RendererItem]
        A list of RendererItem objects.
    deps: Dict[str, List[RendererItem]]
        A dictionary mapping the key of a RendererItem to a list of
        RendererItem that this step depends on.
    step_status: Dict[str, RendererItemStatus], defaults to None.
        A dictionary mapping the key of a RendererItem to its current
        status.
    """

    def __init__(
        self,
        pipeline: "Pipeline" = None,
        pipeline_run: "PipelineRun" = None,
        renderer: PipelineRenderer = None,
    ):
        """Initialize a PipelineVisualizer object.

        Parameters
        ----------
        pipeline: Pipeline
            Pipeline instance.
        pipeline_run: PipelineRun
            PipelineRun instance.
        renderer: PipelineRenderer
            Renderer used to visualize pipeline in text or graph.
        """
        self.pipeline = None
        self.pipeline_run = None
        self.renderer = None
        self.steps = []
        self.deps = {}
        self.step_status = {}

        if pipeline:
            self.with_pipeline(pipeline)
        if pipeline_run:
            self.with_pipeline_run(pipeline_run)
        if renderer:
            self.with_renderer(renderer)

    def with_renderer(self, value: PipelineRenderer) -> "PipelineVisualizer":
        """
        Add renderer to visualize pipeline.

        Parameters
        ----------
        value: object
            Renderer used to visualize pipeline in text or graph.

        Returns
        -------
        PipelineVisualizer
            The PipelineVisualizer instance.

        Raises
        ------
        PipelineVisualizerError
            If `renderer` not specified.
        """
        if not value:
            raise PipelineVisualizerError("The `renderer` must be specified.")

        self.renderer = value
        return self

    def with_pipeline(self, value: "Pipeline") -> "PipelineVisualizer":
        """
        Adds a Pipeline instance to be rendered.

        Parameters
        ----------
        value: Pipeline
            Pipeline instance.

        Returns
        -------
        PipelineVisualizer
            The PipelineVisualizer instance.

        Raises
        ------
        PipelineVisualizerError
            If `pipeline` not specified.
        """
        if not value:
            raise PipelineVisualizerError("The `pipeline` must be specified.")

        self.pipeline = value
        pipeline_render_item = RendererItem(
            name=self.pipeline.name, kind=self.pipeline.kind, spec=self.pipeline
        )
        self.steps = [pipeline_render_item]
        self.deps = {pipeline_render_item.key: []}

        if self.pipeline.step_details:
            render_item_map = {
                step.name: RendererItem(name=step.name, kind=step.kind, spec=step)
                for step in self.pipeline.step_details
            }

            for step in self.pipeline.step_details:
                self.steps.append(render_item_map[step.name])
                if step.depends_on:
                    depends_on = [
                        render_item_map[step_name] for step_name in step.depends_on
                    ]
                else:
                    depends_on = [pipeline_render_item]

                self.deps[render_item_map[step.name].key] = depends_on

        return self

    def with_pipeline_run(self, value: "PipelineRun") -> "PipelineVisualizer":
        """
        Adds a PipelineRun instance to be rendered.

        Parameters
        ----------
        value: PipelineRun
            PipelineRun instance.

        Returns
        -------
        PipelineVisualizer
            The PipelineVisualizer instance.

        Raises
        ------
        PipelineVisualizerError
            If `pipeline run` not specified.
        """
        if not value:
            raise PipelineVisualizerError("The `pipeline run` must be specified.")

        self.pipeline_run = value
        self.step_status = {}

        if self.pipeline_run:
            render_item_status = RendererItemStatus.from_pipeline_run(self.pipeline_run)
            self.step_status[render_item_status.key] = render_item_status
            if self.pipeline_run.step_runs:
                for step_run in self.pipeline_run.step_runs:
                    render_item_status = RendererItemStatus.from_pipeline_step_run(
                        step_run
                    )
                    self.step_status[render_item_status.key] = render_item_status

        return self

    def render(self, rankdir: str = GraphOrientation.TOP_BOTTOM):
        """
        Renders pipeline step status.

        Parameters
        ----------
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.

        Returns
        -------
        None

        Raises
        ------
        PipelineVisualizerError
            If `pipeline` or `renderer` not specified.
        """
        if not (self.steps and self.deps and self.renderer):
            raise PipelineVisualizerError("The `pipeline` must be specified.")

        self.renderer.render(
            steps=self.steps,
            deps=self.deps,
            step_status=self.step_status,
            rankdir=rankdir,
        )

    def to_svg(
        self, uri: str = None, rankdir: str = GraphOrientation.TOP_BOTTOM, **kwargs
    ) -> str:
        """
        Renders pipeline as graph in SVG string.

        Parameters
        ----------
        uri: (string, optional). Defaults to None.
            URI location to save the SVG string.
        rankdir: str, default to "TB".
            Direction of the rendered graph; allowed Values are {"TB", "LR"}.

        Returns
        -------
        str
            Graph in svg format.

        Raises
        ------
        PipelineVisualizerError
            If `pipeline` or `renderer` not specified.
        """
        if not (self.steps and self.deps and self.renderer):
            raise PipelineVisualizerError("The `pipeline` must be specified.")

        return self.renderer.save_to(
            steps=self.steps,
            deps=self.deps,
            step_status=self.step_status,
            rankdir=rankdir,
            uri=uri,
            format="svg",
            **kwargs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import bisect
from collections import defaultdict

from sklearn.base import TransformerMixin
from sklearn.preprocessing import LabelEncoder


class DataFrameLabelEncoder(TransformerMixin):
    """
    Label encoder for pandas.dataframe. dask.dataframe.core.DataFrame
    """

    def __init__(self):
        self.label_encoders = defaultdict(LabelEncoder)

    def fit(self, X):
        """
        Fits a DataFrameLAbelEncoder.
        """
        for column in X.columns:
            if X[column].dtype.name in ["object", "category"]:
                X[column] = X[column].astype(str)
                self.label_encoders[column] = LabelEncoder()
                self.label_encoders[column].fit(X[column])
                label_encoder_classes_ = [
                    str(class_)
                    for class_ in self.label_encoders[column].classes_.tolist()
                ]
                bisect.insort_left(label_encoder_classes_, "unknown")
                self.label_encoders[column].classes_ = label_encoder_classes_
        return self

    def transform(self, X):
        """
        Transforms a dataset using the DataFrameLAbelEncoder.
        """
        categorical_columns = list(self.label_encoders.keys())
        if len(categorical_columns) == 0:
            return X

        def _label_encode_with_unknown(name, series):
            return self.label_encoders[name].transform(
                series.map(lambda x: str(x)).map(
                    lambda x: x
                    if x in self.label_encoders[name].classes_
                    else "unknown"
                )
            )

        X[categorical_columns] = X[categorical_columns].apply(
            lambda series: _label_encode_with_unknown(series.name, series)
        )
        return X



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.dataset.dataset_with_target import ADSDatasetWithTarget
from ads.type_discovery.typed_feature import DateTimeTypedFeature


class ForecastingDataset(ADSDatasetWithTarget):
    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        # index on target
        if isinstance(target, DateTimeTypedFeature):
            df = df.set_index(target)
        ADSDatasetWithTarget.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )

    def select_best_features(self, score_func=None, k=12):
        """
        Not yet implemented
        """
        raise NotImplementedError(
            "Feature selection for forecasting dataset is not yet supported"
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import os
import tempfile
import json
import shlex
from getopt import gnu_getopt

from ads.opctl.constants import (
    ADS_DATAFLOW_CONFIG_FILE_NAME,
    DEFAULT_ADS_CONFIG_FOLDER,
)
from ads.jobs import Job, DataFlow, DataFlowRuntime, DataFlowRun
from ads.jobs.utils import get_dataflow_config
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


def dataflow(line, cell=None):
    opts, args = gnu_getopt(
        shlex.split(line),
        "f:a:c:t:n:hwo",
        longopts=[
            "filename=",
            "archive=",
            "config=",
            "help=",
            "watch=",
            "log-type=",
            "num-lines=",
            "overwrite=",
        ],
    )
    opts_dict = {k: v for k, v in opts}
    if len(args) == 0 and ("-h" in opts_dict or "--help" in opts_dict):
        print(
            f"Run `dataflow run -h` or `dataflow log -h` to see options for subcommands."
        )
        return
    if args[0] not in ("run", "log"):
        print(
            f"`dataflow` expects subcommand `run` or `log`. Use `dataflow log` or `dataflow run` with options."
        )
        return

    if args[0] == "run":
        dataflow_run(opts_dict, args, cell)
    else:

        dataflow_log(opts_dict, args)


def dataflow_run(options, args, cell):
    if "-h" in options or "--help" in options:
        print(
            f"""
Usage: dataflow run [OPTIONS] -- [ARGS]
Options:
    -f, --filename, optional, filename to save the script to. default is "script.py".
    -a, --archive, optional, uri to archive zip.
    -c, --config, optional, configuration passed to dataflow. by default loads from {os.path.join(DEFAULT_ADS_CONFIG_FOLDER, ADS_DATAFLOW_CONFIG_FILE_NAME)}.
    -w, --watch, optional, a flag indicating that dataflow run will be watched after submission.
    -o, --overwrite, optional, overwrite file in object storage when uploading script or archive.zip.
    -h, show this help message.
Args:
    arguments to pass to script.
        """
        )
        return
    if "-c" not in options and "--config" not in options:
        dataflow_config = get_dataflow_config()
    else:
        if "-c" in options:
            dataflow_config = json.loads(options["-c"])
        if "--config" in options:
            dataflow_config = json.loads(options["--config"])
    if "-f" in options:
        script_name = options["-f"]
    elif "--file" in options:
        script_name = options["--file"]
    else:
        script_name = "script.py"
    if "-a" in options:
        archive_name = options["-a"]
    elif "--archive" in options:
        archive_name = options["--archive"]
    else:
        archive_name = None
    with tempfile.TemporaryDirectory() as td:
        file = os.path.join(td, script_name)
        with open(file, "w") as f:
            f.write(cell)
        rt_spec = {
            "scriptPathURI": file,
            "scriptBucket": dataflow_config.pop("script_bucket"),
        }
        if len(args) > 1:
            rt_spec["args"] = args[1:]
        if archive_name:
            rt_spec["archiveUri"] = archive_name
            rt_spec["archiveBucket"] = dataflow_config.pop("archive_bucket")
        rt = DataFlowRuntime(rt_spec)
        infra = DataFlow(spec=dataflow_config)
        if "-o" in options or "--overwrite" in options:
            df = infra.create(rt, overwrite=True)
        else:
            df = infra.create(rt)
        print("DataFlow App ID", df.id)
        df_run = df.run()
        print("DataFlow Run ID", df_run.id)
        print("DataFlow Run Page", df_run.run_details_link)
        if "-w" in options or "--watch" in options:
            df_run.watch()
            print(df_run.logs.application.stdout.tail())


def dataflow_log(options, args):
    if "-h" in options or "--help" in options:
        print(
            f"""
Usage: dataflow log [OPTIONS] [RUN_ID]
Options:
    -t, --log-type, optional, should be one of application, driver, executor. default is "application".
    -n, --num-lines, optional, show last `n` lines of the log
    -h, show this help message.
        """
        )
        return
    log_type = (
        options.get("-t", None) or options.get("--log-type", None) or "application"
    )
    if log_type not in ("application", "driver", "executor"):
        print("Log type should be one of application, driver, executor.")
        return
    n = options.get("-n", None) or options.get("--num-lines", None)
    n = int(n) if (n and len(n) > 0) else None
    if len(args) < 2:
        raise ValueError("DataFlow Run ID must be provided.")
    df_run = DataFlowRun.from_ocid(args[1])
    if log_type == "application":
        print(df_run.logs.application.stdout.tail(n=n))
    if log_type == "driver":
        print(df_run.logs.driver.stdout.tail(n=n))
    if log_type == "executor":
        print(df_run.logs.executor.stdout.tail(n=n))


@runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
def load_ipython_extension(ipython):
    ipython.register_magic_function(dataflow, "line_cell")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import importlib
import json
import logging
import os
import runpy
import subprocess
import sys
import time
import traceback
from typing import List
from urllib.parse import urlparse


import oci


CONST_ENV_LOG_LEVEL = "OCI_LOG_LEVEL"
CONST_ENV_WORKING_DIR = "WORKING_DIR"
CONST_ENV_CODE_DIR = "CODE_DIR"
CONST_ENV_JOB_RUN_OCID = "JOB_RUN_OCID"
CONST_ENV_PYTHON_PATH = "PYTHON_PATH"
CONST_ENV_ENTRYPOINT = "CODE_ENTRYPOINT"
CONST_ENV_ENTRY_FUNC = "ENTRY_FUNCTION"
CONST_ENV_OUTPUT_DIR = "OUTPUT_DIR"
CONST_ENV_OUTPUT_URI = "OUTPUT_URI"
CONST_ENV_OCI_RP = "OCI_RESOURCE_PRINCIPAL_VERSION"
CONST_ENV_ADS_IAM = "OCI_IAM_TYPE"
CONST_API_KEY = "api_key"


DEFAULT_CODE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    os.environ.get(CONST_ENV_CODE_DIR, "code"),
)


def set_log_level(the_logger: logging.Logger):
    """Sets the log level of a logger based on the environment variable.
    This will also set the log level of logging.lastResort.

    Parameters
    ----------
    the_logger : logging.Logger
        A logger object.
    """
    # Do nothing if env var is not set or is empty/None
    if not os.environ.get(CONST_ENV_LOG_LEVEL):
        return the_logger
    log_level = os.environ.get(CONST_ENV_LOG_LEVEL)
    try:
        the_logger.setLevel(log_level)
        logging.lastResort.setLevel(log_level)
        the_logger.info(f"Log level set to {log_level}")
    except Exception:
        # Catching all exceptions here
        # Setting log level should not interrupt the job run even if there is an exception.
        the_logger.warning("Failed to set log level.")
        the_logger.debug(traceback.format_exc())
    return the_logger


logger = logging.getLogger(__name__)
set_log_level(logger)


class OCIHelper:
    """Contains helper functions to call OCI APIs"""

    @staticmethod
    def init_oci_client(client_class):
        """Initializes OCI client with API key or Resource Principal.

        Parameters
        ----------
        client_class :
            The class of OCI client to be initialized.
        """
        if (
            os.environ.get(CONST_ENV_ADS_IAM, "").lower() == CONST_API_KEY
            or CONST_ENV_OCI_RP not in os.environ
        ):
            logger.info("Initializing %s with API Key...", {client_class.__name__})
            client = client_class(
                oci.config.from_file(
                    file_location=os.environ.get(
                        "OCI_CONFIG_LOCATION", oci.config.DEFAULT_LOCATION
                    ),
                    profile_name=os.environ.get(
                        "OCI_CONFIG_PROFILE", oci.config.DEFAULT_PROFILE
                    ),
                )
            )
        else:
            logger.info(
                "Initializing %s with Resource Principal...", client_class.__name__
            )
            client = client_class(
                {}, signer=oci.auth.signers.get_resource_principals_signer()
            )
        return client

    @staticmethod
    def copy_to_oci_object_storage(
        output_dir: str, namespace: str, bucket: str, prefix: str
    ) -> List[str]:
        """Copies the output files to OCI object storage

        Parameters
        ----------
        output_dir : str
            Path of the output directory containing files to be copied.
        namespace : str
            Namespace of the object storage location.
        bucket : str
            Bucket name of the object storage location.
        prefix : str
            Prefix (path) of the object storage location.

        Returns
        -------
        list
            A list of URIs for files that are copied to output_uri
        """
        client = OCIHelper.init_oci_client(oci.object_storage.ObjectStorageClient)

        if not prefix:
            prefix = ""
        prefix = prefix.strip("/")

        copied = []
        for path, _, files in os.walk(output_dir):
            for name in files:
                file_path = os.path.join(path, name)
                # Get the relative path of the file to keep the directory structure
                relative_path = os.path.relpath(file_path, output_dir)
                if prefix:
                    file_prefix = os.path.join(prefix, relative_path)
                else:
                    # Save file to bucket root if prefix is empty.
                    file_prefix = relative_path

                logger.debug(
                    "Saving %s to %s@%s/%s",
                    relative_path,
                    bucket,
                    namespace,
                    file_prefix,
                )

                with open(file_path, "rb") as pkf:
                    client.put_object(
                        namespace,
                        bucket,
                        file_prefix,
                        pkf,
                    )
                copied.append(f"{bucket}@{namespace}/{file_prefix}")
        return copied

    @staticmethod
    def substitute_output_uri(output_uri):
        """Expand shell variables of form $var and ${var}.
        Unknown variables are left unchanged.
        """
        try:
            return os.path.expandvars(output_uri)
        except Exception:
            logger.warning("Failed to expand output URI with environment variables.")
            logger.debug(traceback.format_exc())
            # Do nothing if there is an error.
            return output_uri

    @staticmethod
    def copy_outputs(
        output_dir: str = os.environ.get("OUTPUT_DIR"),
        output_uri: str = os.environ.get("OUTPUT_URI"),
    ) -> List[str]:
        """Copies the output files to remote URI.

        No file will be copied if either output_dir or output_uri is empty or None.

        Parameters
        ----------
        output_dir : str
            Path of the output directory containing files to be copied.
            By default, os.environ.get("OUTPUT_DIR")
        output_uri : str
            URI of the object storage path to store the output files.
            By default, os.environ.get("OUTPUT_URI")

        Returns
        -------
        list
            A list of URIs for files that are copied to output_uri
        """
        if not output_dir:
            logger.info("OUTPUT_DIR is not defined. No file is copied.")
            return
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        if not os.path.exists(output_dir):
            logger.error("Output directory %s not found.", output_dir)
            return

        if not output_uri:
            logger.info("OUTPUT_URI is not defined. No file is copied.")
            return
        output_uri = OCIHelper.substitute_output_uri(output_uri)

        logger.debug("Copying files in %s to %s...", output_dir, output_uri)
        parsed = urlparse(output_uri)
        # Only OCI object storage is supported at the moment

        bucket = parsed.username
        namespace = parsed.hostname
        if not bucket or not namespace:
            logger.error(
                "Invalid bucket name or namespace in output URI: %s", output_uri
            )
            logger.error(
                "Output URI should have the format of oci://bucket@namespace/path/to/dir"
            )
            return

        prefix = parsed.path
        return OCIHelper.copy_to_oci_object_storage(
            output_dir, namespace, bucket, prefix
        )


class ArgumentParser:
    """Contains methods for parsing arguments for entry function."""

    def __init__(self, argument_list: list) -> None:
        """Initialize the parser with a list of arguments

        Parameters
        ----------
        argument_list : list
            A list of arguments.
        """
        self.argument_list = argument_list

    @staticmethod
    def decode_arg(val: str):
        """Decodes the value of the argument if it is a JSON payload.

        Parameters
        ----------
        val : str
            The argument value in a string.

        Returns
        -------
        Any
            None, if the val is None.
            String value, if the val is a string but not a JSON payload.
            Otherwise, the object after JSON decoded.
        """
        if val is None:
            return None
        try:
            return json.loads(val)
        except json.decoder.JSONDecodeError:
            return val

    @staticmethod
    def join_values(value_list: list):
        """Joins the values of a keyword argument.

        Parameters
        ----------
        value_list : list
            Values in a list of strings.

        Returns
        -------
        str or None
            The value of the argument as a string.
        """
        if value_list:
            return " ".join(value_list)
        return None

    def parse(self):
        """Parses the arguments

        Returns
        -------
        (list, dict)
            A tuple of positional arguments (list) and keyword arguments (dict).
        """
        args = []
        kwargs = {}
        parsing_kwargs = False
        key = None
        val = []
        for arg in self.argument_list:
            arg = str(arg)
            if len(arg) > 2 and arg.startswith("--"):
                if key:
                    # Save previous key and val
                    kwargs[key] = self.join_values(val)
                parsing_kwargs = True
                key = arg[2:]
                # Reset val
                val = []
            elif parsing_kwargs:
                val.append(arg)
            else:
                args.append(arg)
        # Save the last key and val
        if key:
            kwargs[key] = self.join_values(val)

        args = [self.decode_arg(arg) for arg in args]
        kwargs = {k: self.decode_arg(v) for k, v in kwargs.items()}
        return args, kwargs


class JobRunner:
    def __init__(self, code_dir: str = DEFAULT_CODE_DIR) -> None:
        """Initialize the job runner

        Parameters
        ----------
        code_dir : str
            The path to the directory containing the user code.
        """
        logger.info("Job Run ID is: %s", os.environ.get(CONST_ENV_JOB_RUN_OCID))
        self.code_dir = code_dir

    @staticmethod
    def run_command(
        command: str, activate_conda: bool = False, level: int = logging.INFO
    ) -> int:
        """Runs a shell command and logs the outputs with specific log level.

        Parameters
        ----------
        command : str
            The shell command
        activate_conda : bool, optional
            Indicate if conda environment should be activated for running the command, by default False
        level : int, optional
            Logging level for the command outputs, by default logging.INFO

        Returns
        -------
        int
            The return code of the command.
        """
        logger.debug(">>> %s", command)
        if activate_conda:
            # Conda activate
            # https://docs.conda.io/projects/conda/en/latest/release-notes.html#id241
            conda_prefix = sys.executable.split("/bin/python", 1)[0]
            cmd = (
                "CONDA_BASE=$(conda info --base) && "
                + "source $CONDA_BASE/etc/profile.d/conda.sh && "
                + f"conda activate {conda_prefix}; "
                + command
            )
        else:
            cmd = command
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=os.environ.copy(),
            shell=True,
        )
        # Steam the outputs
        while True:
            output = process.stdout.readline()
            if process.poll() is not None and output == b"":
                break
            if output:
                # logging will flush outputs by default
                logger.log(level=level, msg=output.decode().strip())
            time.sleep(0.5)
        return process.returncode

    def conda_unpack(self):
        if self.run_command("conda-unpack"):
            logger.info("conda-unpack exits with non-zero return code.")
        return self

    def set_working_dir(self, working_dir: str = os.environ.get(CONST_ENV_WORKING_DIR)):
        """Sets the working directory for the job run.

        Parameters
        ----------
        working_dir : str, optional
            Working directory, by default os.environ.get("WORKING_DIR")
            If working_dir is not set, the working dir will be set to the code dir.
            If working_dir is a relative path, it will be joined with the code dir.

        """
        if working_dir:
            working_dir = os.path.join(self.code_dir, working_dir)
        else:
            working_dir = self.code_dir
        os.chdir(working_dir)
        # Add working dir to sys.path
        if working_dir not in sys.path:
            sys.path.append(working_dir)
        return self

    def setup_python_path(
        self, python_paths: str = os.environ.get(CONST_ENV_PYTHON_PATH, "")
    ):
        """Adds additional python paths.
        Relative paths are expanded based on the current working directory.
        This method should be called after setting the working directory (if needed).

        Parameters
        ----------
        python_paths : str
            Additional python paths to be added to sys.path,
            by default, os.environ.get("PYTHON_PATH", "")
            Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac.

        """
        path_list = python_paths.split(os.pathsep)
        path_list.append(self.code_dir)
        for path in path_list:
            python_path = os.path.abspath(os.path.expanduser(path))
            if python_path not in sys.path:
                sys.path.append(python_path)
        logger.debug("Python Path: %s", sys.path)
        return self

    def _run_function(self, module_path: str, entry_function: str, argv: list):
        """Runs the entry function in module specified by module path.

        Parameters
        ----------
        module_path : str
            The path to the module containing the entry function.
        entry_function : str
            The name of the entry function.
        argv : list
            Argument list from command line.
        This list will be parsed into positional arguments and keyword arguments.
        """
        spec = importlib.util.spec_from_file_location("module", module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        method = getattr(module, entry_function)

        args, kwargs = ArgumentParser(argv).parse()

        logger.debug(
            "Invoking method: %s with args: %s, kwargs: %s",
            entry_function,
            args,
            kwargs,
        )
        method(*args, **kwargs)

    def run(
        self,
        entrypoint: str = os.environ.get(CONST_ENV_ENTRYPOINT),
        entry_function: str = os.environ.get(CONST_ENV_ENTRY_FUNC),
    ):
        """Runs the user code

        Parameters
        ----------
        entrypoint : str
            Path to the file serve as the entrypoint,
            by default, os.environ.get("CODE_ENTRYPOINT")

        entry_function : str, optional
            Name of the function in the entrypoint,
            by default, os.environ.get("ENTRY_FUNCTION").
            If this is not set, the entrypoint will be run as a python script.

        """
        if not entrypoint:
            raise ValueError(f"Invalid entrypoint: {str(entrypoint)}")
        if entry_function:
            logger.info("Running function: %s in %s", entry_function, entrypoint)
            self._run_function(entrypoint, entry_function, sys.argv[1:])
        elif entrypoint.endswith(".ipynb"):
            from driver_notebook import run_notebook

            logger.info("Running notebook: %s", entrypoint)
            # Pass in the absolute path to make sure the working dir is notebook directory
            run_notebook(os.path.abspath(os.path.expanduser(entrypoint)))
        else:
            logger.info("Running script: %s", entrypoint)
            runpy.run_path(entrypoint, run_name="__main__")
        logger.info("Job run completed.")
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
This exploratory data analysis (EDA) Mixin is used in the ADS accessor for the Pandas Series.
The series of purpose-driven methods enable the data scientist to complete univariate analysis.

From the accessor we have access to the pandas object the user is interacting with as well as
corresponding list of feature types.
"""

import pandas as pd
import matplotlib.pyplot as plt


class EDAMixinSeries:
    def feature_stat(self) -> pd.DataFrame:
        """Summary statistics Dataframe provided.

        This returns feature stats on series using FeatureType summary method.

        Examples
        --------
        >>> df = pd.read_csv('~/advanced-ds/tests/vor_datasets/vor_titanic.csv')
        >>> df['Cabin'].ads.feature_stat()
            Metric      Value
        0	count	    891
        1	unqiue	    147
        2	missing	    687

        Returns
        -------
        :class:`pandas.DataFrame`
            Dataframe with 2 columns and rows for different metric values

        """
        for feature_type in self._feature_type:
            if hasattr(feature_type, "feature_stat"):
                stat = feature_type.feature_stat(self._obj).reset_index()
                stat.columns = ["Metric", "Value"]
                return stat
        return None

    def feature_plot(self) -> plt.Axes:
        """For the series generate a summary plot based on the most relevant feature type.

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
           Plot object for the series based on the most relevant feature type.
        """
        for feature_type in self._feature_type:
            if hasattr(feature_type, "feature_plot"):
                return feature_type.feature_plot(self._obj)
        return None

    def warning(self) -> pd.DataFrame:
        """Generates a data frame that lists feature specific warnings.

        Returns
        -------
        :class:`pandas.DataFrame`
            The list of feature specific warnings.

        Examples
        --------
        >>> df["Age"].ads.warning()
          Feature Type       Warning               Message         Metric      Value
         ---------------------------------------------------------------------------
        0   continuous         Zeros      Age has 38 zeros          Count         38
        1   continuous         Zeros   Age has 12.2% zeros     Percentage      12.2%
        """
        common_columns = ["Feature Type", "Warning", "Message", "Metric", "Value"]
        result_df = pd.DataFrame((), columns=common_columns)
        for feature_type in self._feature_type:
            if hasattr(feature_type, "warning"):
                warning_df = feature_type.warning(self._obj)
                if warning_df is not None:
                    warning_df.insert(0, "Feature Type", feature_type.name)
                    result_df = result_df.append(warning_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import absolute_import, print_function

import base64
from io import BytesIO
from typing import Dict, List, Union
import requests
import json
import numpy as np
import pandas as pd

DEFAULT_CONTENT_TYPE_JSON = "application/json"
DEFAULT_CONTENT_TYPE_BYTES = "application/octet-stream"


class InputDataSerializer:
    """[An internal class]
    Defines the contract for input data

    """

    def __init__(
        self,
        data: Union[
            Dict,
            str,
            List,
            np.ndarray,
            pd.core.series.Series,
            pd.core.frame.DataFrame,
        ],
        data_type=None,
    ):
        """
        Parameters
        ----------
        data: Union[Dict, str, list, numpy.ndarray, pd.core.series.Series,
        pd.core.frame.DataFrame]
            Data expected by the model deployment predict API.
        data_type: Any, defaults to None.
            Type of the data. If not provided, it will be checked against data.

        """
        if not data_type:
            data_type = type(data)
        if isinstance(data, np.ndarray):
            np_bytes = BytesIO()
            np.save(np_bytes, data, allow_pickle=True)
            data = base64.b64encode(np_bytes.getvalue()).decode("utf-8")
        elif isinstance(data, pd.core.series.Series):
            data = data.tolist()
        elif isinstance(data, pd.core.frame.DataFrame):
            data = data.to_json()
        elif (
            isinstance(data, dict)
            or isinstance(data, str)
            or isinstance(data, list)
            or isinstance(data, tuple)
            or isinstance(data, bytes)
        ):
            pass
        else:
            raise TypeError(
                "The supported data types are Dict, str, list, "
                "numpy.ndarray, pd.core.series.Series, "
                "pd.core.frame.DataFrame, bytes. Please "
                "convert to the supported data types first. "
            )

        self._data = data
        self._data_type = str(data_type)

    @property
    def data(self):
        return self._data

    @property
    def data_type(self):
        return self._data_type

    def to_dict(self):
        return {
            "data": self._data,
            "data_type": self._data_type,
        }

    def is_bytes(self):
        return "bytes" in self.data_type

    def send(self, endpoint: str, dry_run: bool = False, **kwargs):
        headers = dict()
        if self.is_bytes():
            headers["Content-Type"] = (
                kwargs.get("content_type") or DEFAULT_CONTENT_TYPE_BYTES
            )
            request_kwargs = {"data": self.data}  # should pass bytes when using data
        else:
            headers["Content-Type"] = (
                kwargs.get("content_type") or DEFAULT_CONTENT_TYPE_JSON
            )
            request_kwargs = {"json": self.to_dict()}
        request_kwargs["headers"] = headers

        if dry_run:
            request_kwargs["headers"]["Accept"] = "*/*"
            req = requests.Request("POST", endpoint, **request_kwargs).prepare()
            if self.is_bytes():
                return req.body
            return json.loads(req.body)
        else:
            request_kwargs["auth"] = kwargs.get("signer")
            return requests.post(endpoint, **request_kwargs).json()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

from typing import Any, Dict, Sequence, Union

CategoricalChoiceType = Union[None, bool, int, float, str]


class Distribution:
    """Defines the abstract base class for hyperparameter search distributions"""

    def __init__(self, dist):
        self._dist = dist

    def get_distribution(self):
        """Returns the distribution"""
        return self._dist

    def __repr__(self) -> str:

        kwargs = ", ".join(
            "{}={}".format(k, v)
            for k, v in sorted(self.__dict__.items(), reverse=True)
            if k != "_dist"
        )

        return "{}({})".format(self.__class__.__name__, kwargs)


class DiscreteUniformDistribution(Distribution):
    """
    A discretized uniform distribution in the linear domain.

    .. note::
        If the range :math:`[\\mathsf{low}, \\mathsf{high}]` is not divisible by :math:`q`,
        :math:`\\mathsf{high}` will be replaced with the maximum of :math:`k q + \\mathsf{low}
        \\lt \\mathsf{high}`, where :math:`k` is an integer.

    Parameters
    ----------
        low: float
            Lower endpoint of the range of the distribution. `low` is included in the range.
        high: float
            Upper endpoint of the range of the distribution. `high` is included in the range.
        step: float
            A discretization step.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, low: float, high: float, step: float):
        self.low = low
        self.high = high
        self.q = step
        dist = optuna.distributions.DiscreteUniformDistribution(
            low=low, high=high, q=step
        )
        super().__init__(dist)

    def __repr__(self) -> str:

        key_value_pairs = []
        for key in ["low", "high", "q"]:
            if key == "q":
                key_value_pairs.append("{}={}".format("step", self.__dict__[key]))
            else:
                key_value_pairs.append("{}={}".format(key, self.__dict__[key]))

        kwargs = ", ".join(key_value_pairs)

        return "{}({})".format(self.__class__.__name__, kwargs)


class CategoricalDistribution(Distribution):
    """
    A categorical distribution.

    Parameters
    ----------
    choices:
        Parameter value candidates. It is recommended to restrict the types of the choices
        to the following: :obj:`None`, :class:`bool`, :class:`int`, :class:`float`
        and :class:`str`.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, choices: Sequence[CategoricalChoiceType]):
        self.choices = choices
        dist = optuna.distributions.CategoricalDistribution(choices=choices)
        super().__init__(dist)


class IntLogUniformDistribution(Distribution):
    """A uniform distribution on integers in the log domain.

    Parameters
    ----------
        low:
            Lower endpoint of the range of the distribution. `low` is included in the range.
        high:
            Upper endpoint of the range of the distribution. `high` is included in the range.
        step:
            A step for spacing between values.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, low: float, high: float, step: float = 1):
        self.low = low
        self.high = high
        self.step = step
        dist = optuna.distributions.IntLogUniformDistribution(
            low=low, high=high, step=step
        )
        super().__init__(dist)

    def __repr__(self) -> str:
        key_value_pairs = []
        for key in ["low", "high", "step"]:
            key_value_pairs.append("{}={}".format(key, self.__dict__[key]))
        kwargs = ", ".join(key_value_pairs)

        return "{}({})".format(self.__class__.__name__, kwargs)


class IntUniformDistribution(Distribution):
    """
    A uniform distribution on integers.

    .. note::
        If the range :math:`[\\mathsf{low}, \\mathsf{high}]` is not divisible by
        :math:`\\mathsf{step}`, :math:`\\mathsf{high}` will be replaced with the maximum of
        :math:`k \\times \\mathsf{step} + \\mathsf{low} \\lt \\mathsf{high}`, where :math:`k` is
        an integer.

    Parameters
    ----------
        low:
            Lower endpoint of the range of the distribution. `low` is included in the range.
        high:
            Upper endpoint of the range of the distribution. `high` is included in the range.
        step:
            A step for spacing between values.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, low: float, high: float, step: float = 1):
        self.low = low
        self.high = high
        self.step = step
        dist = optuna.distributions.IntUniformDistribution(
            low=low, high=high, step=step
        )
        super().__init__(dist)

    def __repr__(self) -> str:
        key_value_pairs = []
        for key in ["low", "high", "step"]:
            key_value_pairs.append("{}={}".format(key, self.__dict__[key]))
        kwargs = ", ".join(key_value_pairs)

        return "{}({})".format(self.__class__.__name__, kwargs)


class LogUniformDistribution(Distribution):
    """
    A uniform distribution in the log domain.

    Parameters
    ----------
        low:
            Lower endpoint of the range of the distribution. `low` is included in the range.
        high:
            Upper endpoint of the range of the distribution. `high` is excluded from the range.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, low: float, high: float):
        self.low = low
        self.high = high

        dist = optuna.distributions.LogUniformDistribution(low=low, high=high)
        super().__init__(dist)


class UniformDistribution(Distribution):
    """
    A uniform distribution in the linear domain.

    Parameters
    ----------
        low:
            Lower endpoint of the range of the distribution. `low` is included in the range.
        high:
            Upper endpoint of the range of the distribution. `high` is excluded from the range.
    """

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(self, low: float, high: float):
        self.low = low
        self.high = high

        dist = optuna.distributions.UniformDistribution(low=low, high=high)
        super().__init__(dist)


class DistributionEncode(json.JSONEncoder):
    def default(self, dist: Distribution) -> Dict[str, Any]:
        if isinstance(dist, DiscreteUniformDistribution):
            return {
                "ClassName": DiscreteUniformDistribution.__name__,
                "low": dist.low,
                "high": dist.high,
                "q": dist.q,
            }
        if isinstance(dist, CategoricalDistribution):
            return {
                "ClassName": CategoricalDistribution.__name__,
                "choices": dist.choices,
            }
        if isinstance(dist, IntLogUniformDistribution):
            return {
                "ClassName": IntLogUniformDistribution.__name__,
                "low": dist.low,
                "high": dist.high,
                "step": dist.step,
            }
        if isinstance(dist, IntUniformDistribution):
            return {
                "ClassName": IntUniformDistribution.__name__,
                "low": dist.low,
                "high": dist.high,
                "step": dist.step,
            }
        if isinstance(dist, LogUniformDistribution):
            return {
                "ClassName": LogUniformDistribution.__name__,
                "low": dist.low,
                "high": dist.high,
            }
        if isinstance(dist, UniformDistribution):
            return {
                "ClassName": UniformDistribution.__name__,
                "low": dist.low,
                "high": dist.high,
            }
        return json.JSONEncoder.default(self, dist)

    @staticmethod
    def from_json(json_object: Dict[Any, Any]):
        if "ClassName" in json_object.keys():
            if json_object["ClassName"] == DiscreteUniformDistribution.__name__:
                return DiscreteUniformDistribution(
                    low=json_object["low"],
                    high=json_object["high"],
                    step=json_object["step"],
                )

            if json_object["ClassName"] == CategoricalDistribution.__name__:
                return CategoricalDistribution(choices=json_object["choices"])

            if json_object["ClassName"] == IntLogUniformDistribution.__name__:
                return IntLogUniformDistribution(
                    low=json_object["low"],
                    high=json_object["high"],
                    step=json_object["step"],
                )

            if json_object["ClassName"] == IntUniformDistribution.__name__:
                return IntUniformDistribution(
                    low=json_object["low"],
                    high=json_object["high"],
                    step=json_object["step"],
                )

            if json_object["ClassName"] == LogUniformDistribution.__name__:
                return LogUniformDistribution(
                    low=json_object["low"], high=json_object["high"]
                )

            if json_object["ClassName"] == UniformDistribution.__name__:
                return UniformDistribution(
                    low=json_object["low"], high=json_object["high"]
                )
        else:
            return json_object


def encode(o: Distribution) -> str:
    """Encodes a distribution to a string

    Parameters
    ----------
        o: :class:`Distribution`
            The distribution to encode

    Returns
    -------
        str (:class:`DistributionEncode`)
            The distribution encoded as a string
    """
    return json.dumps(o, cls=DistributionEncode)


def decode(s: str):
    """Decodes a string to an object

    Parameters
    ----------
        s: str
            The string being decoded to a distribution object

    Returns
    -------
        :class:`Distribution` or :class:`Dict`
            Decoded string
    """
    return json.loads(s, object_hook=DistributionEncode.from_json)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import datetime
import logging
import time
from typing import Dict, Union, List

import oci.logging
import oci.loggingsearch
from ads.common.decorator.utils import class_or_instance_method
from ads.common.oci_mixin import OCIModelMixin, OCIWorkRequestMixin
from ads.common.oci_resource import OCIResource, ResourceNotFoundError


logger = logging.getLogger(__name__)

# Maximum number of log records to be returned by default.
LOG_RECORDS_LIMIT = 100
# The time interval in seconds between sending each request to pull logs
LOG_INTERVAL = 3
MAXIMUM_RETRY_COUNT = 20


class SortOrder:
    ASC = "ASC"
    DESC = "DESC"


class OCILoggingModelMixin(OCIModelMixin, OCIWorkRequestMixin):
    """Base model for representing OCI logging resources managed through oci.logging.LoggingManagementClient.
    This class should not be initialized directly. Use a sub-class (OCILogGroup or OCILog) instead.
    """

    @class_or_instance_method
    def from_name(cls, display_name: str, **kwargs) -> Union["OCILogGroup", "OCILog"]:
        """Obtain an existing OCI logging resource by using its display name.
        OCI log group or log resource requires display name to be unique.

        Parameters
        ----------
        display_name : str
            Display name of the logging resource (e.g. log group)

        Returns
        -------
        An instance of logging resource, e.g. OCILogGroup, or OCILog.

        """
        items = cls.list_resource(**kwargs)
        for item in items:
            if item.display_name == display_name:
                return item
        return None

    @class_or_instance_method
    def init_client(cls, **kwargs) -> oci.logging.LoggingManagementClient:
        """Initialize OCI client"""
        return cls._init_client(client=oci.logging.LoggingManagementClient, **kwargs)

    @property
    def client(self) -> oci.logging.LoggingManagementClient:
        """OCI logging management client"""
        return super().client

    def create_async(self):
        """Creates the OCI logging resource asynchronously.
        Sub-class should implement this method with OCI Python SDK and return the response from the OCI PythonSDK.

        """
        raise NotImplementedError()

    def create(self):
        """Creates a new resource with OCI service synchronously.
        This method will wait for the resource creation to be succeeded or failed.

        Each sub-class should implement the create_async() method with the corresponding method in OCI SDK
            to create the resource.

        Raises
        ------
        NotImplementedError
            when user called create but the create_async() method is not implemented.

        oci.exceptions.RequestException
            when there is an error creating the resource with OCI request.

        """
        wait_for_states = ("SUCCEEDED", "FAILED")
        success_state = "SUCCEEDED"

        res = self.create_async()
        # Wait for the work request to be completed.
        if "opc-work-request-id" in res.headers:
            opc_work_request_id = res.headers.get("opc-work-request-id")
            res_work_request = self.wait_for_work_request(
                opc_work_request_id,
                wait_for_state=wait_for_states,
                wait_interval_seconds=1,
            )
            # Raise an error if the failed to create the resource.
            if res_work_request.data.status != success_state:
                raise oci.exceptions.RequestException(
                    f"Failed to create {self.__class__.__name__}.\n" + str(res.data)
                )
            self.id = res_work_request.data.resources[0].identifier
            logger.debug("Created %s: %s", self.__class__.__name__, self.id)

        else:
            # This will likely never happen as OCI SDK will raise an error if the HTTP request is not successful.
            raise oci.exceptions.RequestException(
                f"opc-work-request-id not found in response headers: {res.headers}"
            )
        self.sync()
        return self


class OCILogGroup(OCILoggingModelMixin, oci.logging.models.LogGroup):
    """Represents the OCI Log Group resource.

    Using ``OCILogGroup`` to create a new log group.
    OCI requires display_name to be unique and it cannot contain space.
    >>> log_group = OCILogGroup(display_name="My_Log_Group").create()
    Once created, access the OCID and other properties
    >>> log_group.id # The OCID is None before the log is created.
    >>> None
    Create a log resource within the log group
    >>> log_group.id # OCID will be available once the log group is created
    Access the property
    >>> log_group.display_name
    Create logs within the log group
    >>> log = log_group.create_log("My custom access log")
    >>> log_group.create_log("My custom prediction log")
    List the logs in a log group. The following line will return a list of OCILog objects.
    >>> logs = log_group.list_logs()
    Delete the resource
    >>> log_group.delete()
    """

    def create_async(self):
        """Creates a new LogGroup asynchronously with OCI logging service"""
        self.load_properties_from_env()
        return self.client.create_log_group(
            self.to_oci_model(oci.logging.models.CreateLogGroupDetails)
        )

    def create_log(self, display_name: str, **kwargs):
        """Create a log (OCI resource) within the log group.

        Parameters
        ----------
        display_name : str
            The display name of the log
        **kwargs :
            Keyword arguments to be passed into the OCI API for log properties.

        Returns
        -------
        OCILog
            An instance of OCILog

        """
        return OCILog(
            display_name=display_name,
            log_group_id=self.id,
            compartment_id=self.compartment_id,
            **kwargs,
        ).create()

    def list_logs(self, **kwargs) -> list:
        """Lists all logs within the log group.

        Parameters
        ----------
        **kwargs :
            keyword arguments for filtering the results.
            They are passed into oci.logging.LoggingManagementClient.list_logs()

        Returns
        -------
        list
            A list of OCILog

        """
        items = oci.pagination.list_call_get_all_results(
            self.client.list_logs, self.id, **kwargs
        ).data
        return [OCILog.from_oci_model(item) for item in items]

    def delete(self):
        """Deletes the log group and the logs in the log group."""
        logs = self.list_logs()
        for log in logs:
            logger.debug("Deleting OCI Log: %s", log.id)
            log.delete()
        logger.debug("Deleting OCI log group: %s", self.id)
        self.client.delete_log_group(self.id)
        return self


class OCILog(OCILoggingModelMixin, oci.logging.models.Log):
    """Represents the OCI Log resource.

    Usage: (OCI requires display_name to be unique and it cannot contain space)
    >>> log = OCILog(display_name="My_Log", log_group_id=LOG_GROUP_ID).create()
    Usually it is better to create a log using the create_log() method in OCILogGroup.
    >>> log.delete() # Delete the resource
    Get a log object from OCID
    >>> oci_log = OCILog.from_ocid("LOG_OCID_HERE")
    Stream the logs from an OCI Data Science Job run to stdout
    >>> oci_log.stream(source="JOB_RUN_OCID_HERE")
    Gets the most recent 10 logs
    >>> oci_log.tail(10)
    """

    def __init__(self, log_type: str = "CUSTOM", **kwargs) -> None:
        """Initializes an OCI log model locally.
        The resource is not created in OCI until the create() or create_async() method is called.
        """
        if "logType" not in kwargs:
            kwargs["log_type"] = log_type
        super().__init__(**kwargs)
        self._search_client = None
        self.source = kwargs.get("source", None)
        self.annotation = kwargs.get("annotation", None)

    @staticmethod
    def _get_log_group_id(log_ocid, compartment_ocid: str = None) -> str:
        if not compartment_ocid:
            compartment_ocid = OCIResource.get_compartment_id(log_ocid)
        log_groups = OCILogGroup.list_resource(compartment_id=compartment_ocid)
        for log_group in log_groups:
            oci_logs = log_group.list_logs()
            for oci_log in oci_logs:
                if oci_log.id == log_ocid:
                    return oci_log.log_group_id
        raise ResourceNotFoundError(f"Log not found. OCID={log_ocid}")

    @class_or_instance_method
    def _get(cls, ocid) -> oci.logging.models.Log:
        """Gets the OCI Log data (properties)

        OCI API requires both Log OCID and Log Group OCID to get the log properties.
        This method will lookup the log group ID in the same compartment as the log.
        The look up process involves multiple API calls.
        To avoid the lookup, initialize the OCILog with both Log OCID and Log Group OCID,
        for example: log = OCILog(id=LOG_OCID, log_group_id=LOG_GROUP_OCID)

        Parameters
        ----------
        ocid : str
            OCID of the OCI log resource

        Returns
        -------
        oci.logging.models.Log
            An oci.logging.models.Log object containing the properties of the log resource
        """
        log_group_id = cls._get_log_group_id(ocid)
        return cls().client.get_log(log_group_id, ocid).data

    def create_async(self):
        """Creates a new Log with OCI logging service"""
        self.load_properties_from_env()
        return self.client.create_log(
            self.log_group_id, self.to_oci_model(oci.logging.models.CreateLogDetails)
        )

    def sync(self, **kwargs) -> None:
        """Refreshes the properties of the Log model
        OCI requires both Log OCID and Log group OCID to get the Log model.

        This method override the sync() method from OCIMixin to improve performance.
        """
        if not self.log_group_id:
            self.log_group_id = self._get_log_group_id(self.id, self.compartment_id)
        self.update_from_oci_model(
            self.client.get_log(self.log_group_id, self.id).data, **kwargs
        )
        return self

    def delete(self):
        """Delete the log"""
        self.client.delete_log(self.log_group_id, self.id)
        return self

    @property
    def search_client(self):
        """OCI logging search client."""
        if not self._search_client:
            self._search_client = self._init_client(
                client=oci.loggingsearch.LogSearchClient
            )
        return self._search_client

    @staticmethod
    def format_datetime(dt: datetime.datetime) -> str:
        """Converts datetime object to RFC3339 date time format in string

        Parameters
        ----------
        dt: datetime.datetime
            Datetime object to be formated.


        Returns
        -------
        str:
            A timestamp in RFC3339 format.

        """
        return dt.isoformat()[:-3] + "Z"

    @staticmethod
    def _to_utc_native(dt: datetime.datetime) -> str:
        """Converts offset-aware datetime to offset-native UTC time.

        Parameters
        ----------
        dt : datetime.datetime
            A datetime object to be converted.
            If dt is offset native, it will be returned as is.

        Returns
        -------
        datetime.datetime
            Offset-native datetime
        """
        if dt.tzinfo:
            return dt.astimezone(datetime.timezone.utc).replace(tzinfo=None)
        return dt

    def _search_logs(
        self,
        time_start: datetime.datetime,
        time_end: datetime.datetime,
        search_query: str,
    ) -> List[oci.loggingsearch.models.SearchResult]:
        """Search logs from OCI logging service

        Due to the limitation on the Log search API.
        This method cannot return more than 1000 logs within 1000 microseconds.

        Parameters
        ----------
        time_start : datetime.datetime
            Starting UTC time for the log search
        time_end : datetime.datetime
            Ending UTC time for the log search
        search_query : str
            Search query following the Logging Query Language Specification.
            https://docs.oracle.com/en-us/iaas/Content/Logging/Reference/query_language_specification.htm

        Returns
        -------
        List[oci.loggingsearch.models.SearchResult]
            A list of SearchResult objects
        """
        # The log search API has a default limit of 100 results per request
        # This limit can be set to 1,000
        # https://docs.oracle.com/en-us/iaas/api/#/en/logging-search/20190909/SearchResult/SearchLogs

        LIMIT_PER_REQUEST = 1000
        logger.debug("Requesting logs between %s and %s ...", time_start, time_end)
        search_details = oci.loggingsearch.models.SearchLogsDetails(
            time_start=self.format_datetime(time_start),
            time_end=self.format_datetime(time_end),
            search_query=search_query,
        )
        response = self.search_client.search_logs(
            search_details, limit=LIMIT_PER_REQUEST
        )
        records = response.data.results
        logger.debug("%d logs received.", len(records))
        if len(records) >= LIMIT_PER_REQUEST:
            mid = time_start + (time_end - time_start) / 2
            # The log search API used RFC3339 time format.
            # The minimum time unit of RFC3339 format is 1000 microseconds.
            # In the extreme case, if there are over 1000 logs with a 1k-microsecond interval,
            # only 1000 logs can be returned.
            if time_start + datetime.timedelta(microseconds=1000) > mid:
                return records
            return self._search_logs(time_start, mid, search_query) + self._search_logs(
                mid, time_end, search_query
            )
        return records

    def search(
        self,
        source: str = None,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = None,
        sort_by: str = "datetime",
        sort_order: str = "DESC",
        log_filter: str = None,
    ) -> List[oci.loggingsearch.models.SearchResult]:
        """Search logs

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None. No filtering will be performed.
        time_start : datetime.datetime, optional
            Starting UTC time for the query.
            Defaults to None. Logs from the past 24 hours will be returned.
        time_end : datetime.datetime, optional
            Ending UTC time for the query.
            Defaults to None. The current time will be used.
        limit : int, optional
            Maximum number of records to be returned.
            Defaults to None. All logs will be returned.
        sort_by : str, optional
            The field for sorting the logs.
            Defaults to "datetime"
        sort_order : str, optional
            Specify how the records should be sorted. Must be "ASC" or "DESC".
            Defaults to "DESC".
        log_filter : str, optional
            Expression for filtering the logs.
            This will be the WHERE clause of the query.
            Defaults to None.

        Returns
        -------
        List[oci.loggingsearch.models.SearchResult]
            A list of SearchResult objects

        """
        # Default time_start and time_end
        if time_start is None:
            time_start = datetime.datetime.utcnow() - datetime.timedelta(days=1)
        if time_end is None:
            time_end = datetime.datetime.utcnow()

        time_start = self._to_utc_native(time_start)
        time_end = self._to_utc_native(time_end)

        logger.debug("Searching logs between %s and %s ...", time_start, time_end)

        if time_start >= time_end:
            return []

        # Build search query
        search_query = f'SEARCH "{self.compartment_id}/{self.log_group_id}/{self.id}"'
        filters = []

        filter_source = source or self.source
        if filter_source:
            # source = <OCID> is not allowed but source = *<OCID> works
            if filter_source.startswith("ocid1."):
                filter_source = "*" + filter_source
            filters = [f"source = '{filter_source}'"] + filters

        if log_filter:
            filters.append(log_filter)

        if filters:
            search_query += " | WHERE " + " AND ".join(filters)

        if sort_by:
            if not sort_order:
                sort_order = "DESC"
            search_query += f" | SORT BY {sort_by} {sort_order}"
        logger.debug("Log search query: %s", search_query)

        # Save original datetime object before conversion
        orig_time_start = time_start
        # For each search request, time_start cannot be more than 14 days older
        if time_start < time_end - datetime.timedelta(days=14):
            time_start = time_end - datetime.timedelta(days=14)

        results = []

        while not limit or len(results) < limit:
            if time_start + datetime.timedelta(microseconds=1000) > time_end:
                # The log search API used RFC3339 time format.
                # The minimum time unit of RFC3339 format is 1000 microseconds.
                # If the search interval is less than this,
                # there will be InvalidParameter error from OCI.
                records = []
            elif limit:
                logger.debug(
                    "Requesting logs between %s and %s ...", time_start, time_end
                )
                search_details = oci.loggingsearch.models.SearchLogsDetails(
                    # time_start cannot be more than 14 days older
                    time_start=self.format_datetime(time_start),
                    time_end=self.format_datetime(time_end),
                    # https://docs.oracle.com/en-us/iaas/Content/Logging/Reference/query_language_specification.htm
                    # Double quotes must be used for "<log_stream>" after search
                    # Single quotes must be used for the string in <where_expression>
                    search_query=search_query,
                    # is_return_field_info=True
                )
                response = self.search_client.search_logs(
                    search_details, limit=(limit - len(results))
                )
                records = response.data.results
                logger.debug("%d logs received.", len(records))
            else:
                records = self._search_logs(time_start, time_end, search_query)

            results.extend(records)

            if time_start > orig_time_start:
                time_end = time_start
                time_start = time_end - datetime.timedelta(days=14)
                time_start = max(time_start, orig_time_start)
            else:
                break

        return results

    def _search_and_format(
        self,
        source: str = None,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = LOG_RECORDS_LIMIT,
        sort_by: str = "datetime",
        sort_order: str = SortOrder.DESC,
        log_filter: str = None,
    ):
        """Returns the formatted log records.

        Parameters
        ----------
        source: (str, optional). Defaults to None.
            Filter the records by the source field.
        time_start: (datetime.datetime, optional). Defaults to None.
            Starting time for the query.
        time_end: (datetime.datetime, optional). Defaults to None.
            Ending time for the query.
        limit: (int, optional). Defaults to 100.
            Maximum number of records to be returned.
        sort_by: (str, optional). Defaults to "datetime"
            The field for sorting the logs.
        sort_order: (str, optional). Defaults to "DESC".
            The sort order for the log records. Can be "ASC" or "DESC".
        log_filter : (str, optional). Defaults to None.
            Expression for filtering the logs.
            This will be the WHERE clause of the query.

        Returns
        -------
        list
            A list of log records.
            Each log record is a dictionary with the following keys: `id`, `time`, `message`.
        """
        if not time_start:
            time_start = datetime.datetime.utcnow() - datetime.timedelta(days=14)

        logs = self.search(
            source=source,
            time_start=time_start,
            time_end=time_end,
            limit=limit,
            sort_by=sort_by,
            sort_order=sort_order,
            log_filter=log_filter,
        )
        logs = sorted((log.data for log in logs), key=lambda x: x.get("datetime"))
        logs = [log.get("logContent", {}) for log in logs]
        return [
            {
                "id": log.get("id"),
                "message": log.get("data", {}).get("message"),
                "time": log.get("time"),
            }
            for log in logs
        ]

    def tail(
        self,
        source: str = None,
        limit=LOG_RECORDS_LIMIT,
        time_start: datetime.datetime = None,
        log_filter: str = None,
    ) -> List[dict]:
        """Returns the most recent log records.

        Parameters
        ----------
        source: (str, optional). Defaults to None.
            The source field to filter the log records.
        limit: (int, optional). Defaults to 100.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
        time_start: (datetime.datetime, optional)
            Starting time for the log query.
            Defaults to None. Logs up to 14 days from now will be returned.
        log_filter : (str, optional). Defaults to None.
            Expression for filtering the logs.
            This will be the WHERE clause of the query.

        Returns
        -------
        list
            A list of log records.
            Each log record is a dictionary with the following keys: `id`, `time`, `message`.
        """
        return self._search_and_format(
            source=source,
            limit=limit,
            sort_order=SortOrder.DESC,
            time_start=time_start,
            log_filter=log_filter,
        )

    def head(
        self,
        source: str = None,
        limit=LOG_RECORDS_LIMIT,
        time_start: datetime.datetime = None,
    ) -> List[dict]:
        """Returns the preceding log records.

        Parameters
        ----------
        source: (str, optional). Defaults to None.
            The source field to filter the log records.
        limit: (int, optional). Defaults to 100.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
        time_start: (datetime.datetime, optional)
            Starting time for the log query.
            Defaults to None. Logs up to 14 days from now will be returned.

        Returns
        -------
        list
            A list of log records.
            Each log record is a dictionary with the following keys: `id`, `time`, `message`.
        """
        return self._search_and_format(
            source=source, limit=limit, sort_order=SortOrder.ASC, time_start=time_start
        )

    def stream(
        self,
        source: str = None,
        interval: int = LOG_INTERVAL,
        stop_condition: callable = None,
        time_start: datetime.datetime = None,
        log_filter: str = None,
    ):
        """Streams logs to console/terminal until stop_condition() returns true.

        Parameters
        ----------
        source : str
        interval : int
            The time interval between sending each request to pull logs from OCI logging service (Default value = 3)
        stop_condition : callable
            A function to determine if the streaming should stop. (Default value = None)
            The log streaming will stop if the function returns true.
        time_start: datetime.datetime
            Starting time for the log query.
            Defaults to None. Logs up to 14 days from now will be returned.
        log_filter : str, optional
            Expression for filtering the logs.
            This will be the WHERE clause of the query.
            Defaults to None.

        Returns
        -------
        int:
            The number of logs printed.
        """
        # Use a set to store the IDs of the printed messages
        printed = set()
        exception_count = 0
        while True:
            try:
                # Tail the logs from time_start until now
                # Logs may come in after the timestamp.
                # For the next tail() call,
                # use the time that's 3 minutes before now so that we can have some overlaps.
                next_time_start = datetime.datetime.utcnow() - datetime.timedelta(
                    seconds=180
                )
                logs = self.tail(
                    source, limit=None, time_start=time_start, log_filter=log_filter
                )
                # Update time_start if the tail() is successful
                time_start = next_time_start
            except Exception:
                exception_count += 1
                if exception_count > MAXIMUM_RETRY_COUNT:
                    raise
                else:
                    time.sleep(interval)
                    continue
            for log in logs:
                if log.get("id") not in printed:
                    timestamp = log.get("time", "")
                    if timestamp:
                        timestamp = timestamp.split(".")[0].replace("T", " ")
                    else:
                        timestamp = ""
                    print(f"{timestamp} - {log.get('message')}")
                    printed.add(log.get("id"))
            if stop_condition and stop_condition():
                return len(printed)
            time.sleep(interval)


class ConsolidatedLog:
    """Represents the Consolidated OCI Log resource.

    Usage:
    Initialize consolidated log instance for oci_log_one and oci_log_two
    >>> oci_log_one = OCILog(
    >>>     compartment_id=<compartment_id_one>,
    >>>     id=<id_one>,
    >>>     log_group_id=<log_group_id_one>,
    >>>     annotation=<annotation_one>
    >>> )
    >>> oci_log_two = OCILog(
    >>>     compartment_id=<compartment_id_two>,
    >>>     id=<id_two>,
    >>>     log_group_id=<log_group_id_two>,
    >>>     annotation=<annotation_two>
    >>> )
    >>> consolidated_log = ConsolidatedLog(oci_log_one, oci_log_two)
    Stream, sort and annotate the logs from oci_log_one and oci_log_two
    >>> consolidated_log.stream()
    Get the most recent 20 consolidated logs from oci_log_one and oci_log_two
    >>> consolidated_log.tail(limit=20)
    Get the most recent 20 raw logs from oci_log_one and oci_log_two
    >>> consolidated_log.search(limit=20)
    """

    def __init__(self, *args) -> None:
        """Initializes a consolidate log model instance.

        Parameters
        ----------
        args:
            A list of OCILog instance.
        """
        self.logging_instance = []
        for arg in args:
            if isinstance(arg, OCILog):
                self.logging_instance.append(arg)
            else:
                raise ValueError(
                    "Unrecognized type. ConsolidatedLog constructor requires OCILog instances as parameters."
                )

    def stream(
        self,
        source: str = None,
        interval: int = LOG_INTERVAL,
        stop_condition: callable = None,
        time_start: datetime.datetime = None,
        log_filter: str = None,
    ):
        """Streams consolidated logs to console/terminal until stop_condition() returns true.

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        interval : int, optional
            The time interval between sending each request to pull logs from OCI logging service.
            Defaults to 3.
        stop_condition : callable, optional
            A function to determine if the streaming should stop.
            The log streaming will stop if the function returns true.
            Defaults to None.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.
        """
        self._print_log_annotation_message()
        # Use a set to store the IDs of the printed messages
        printed = set()
        exception_count = 0
        while True:
            try:
                # Tail the logs from time_start until now
                # Logs may come in after the timestamp.
                # For the next tail() call,
                # use the time that's 3 minutes before now so that we can have some overlaps.
                next_time_start = datetime.datetime.utcnow() - datetime.timedelta(
                    seconds=180
                )
                logs = self._search_and_format(
                    source=source,
                    limit=None,
                    sort_order=SortOrder.DESC,
                    time_start=time_start,
                    log_filter=log_filter,
                )
                # Update time_start if the tail() is successful
                time_start = next_time_start
            except Exception:
                exception_count += 1
                if exception_count > MAXIMUM_RETRY_COUNT:
                    raise
                else:
                    time.sleep(interval)
                    continue
            for log in logs:
                if log.get("id") not in printed:
                    self._print_log_details(log)
                    printed.add(log.get("id"))
            if stop_condition and stop_condition():
                return len(printed)
            time.sleep(interval)

    def tail(
        self,
        source: str = None,
        limit: int = LOG_RECORDS_LIMIT,
        time_start: datetime.datetime = None,
        log_filter: str = None,
    ) -> None:
        """Returns the most recent consolidated log records.

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        limit : int, optional.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
            Defaults to 100.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.
        """
        self._print(
            self._search_and_format(
                source=source,
                limit=limit,
                sort_order=SortOrder.DESC,
                time_start=time_start,
                log_filter=log_filter,
            )
        )

    def head(
        self,
        source: str = None,
        limit: int = LOG_RECORDS_LIMIT,
        time_start: datetime.datetime = None,
    ) -> None:
        """Returns the preceding consolidated log records.

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        limit : int, optional.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
            Defaults to 100.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        """
        self._print(
            self._search_and_format(
                source=source,
                limit=limit,
                sort_order=SortOrder.ASC,
                time_start=time_start,
            )
        )

    def _print_log_annotation_message(self) -> None:
        message = ""
        for instance in self.logging_instance:
            if instance.annotation:
                message = (
                    f"[{instance.annotation[0].upper()}] - {instance.annotation} log, "
                    + message
                )
        if message:
            print(message[:-2])

    def _print(self, logs: List[Dict]) -> None:
        self._print_log_annotation_message()
        for log in logs:
            self._print_log_details(log)

    @staticmethod
    def _print_log_details(log) -> None:
        annotation = log.get("annotation", "")
        if annotation:
            annotation = annotation[0].upper()
        timestamp = log.get("time", "")
        if timestamp:
            timestamp = timestamp.split(".")[0].replace("T", " ")
        print(f"[{annotation}] - {timestamp} - {log.get('message', '')}")

    def search(
        self,
        source: str = None,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = None,
        sort_by: str = "datetime",
        sort_order: str = SortOrder.DESC,
        log_filter: str = None,
    ) -> List[oci.loggingsearch.models.SearchResult]:
        """Searches raw logs.

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        time_end : datetime.datetime, optional.
            Ending time for the query.
            Defaults to None.
        limit : int, optional.
            Maximum number of records to be returned. All logs will be returned.
            Defaults to None.
        sort_by : str, optional.
            The field for sorting the logs.
            Defaults to "datetime"
        sort_order : str, optional.
            The sort order for the log records. Can be "ASC" or "DESC".
            Defaults to "DESC".
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.

        Returns
        -------
        list
            A list of oci.loggingsearch.models.SearchResult objects
        """
        return self._search_and_format(
            source=source,
            time_start=time_start,
            time_end=time_end,
            limit=limit,
            sort_by=sort_by,
            sort_order=sort_order,
            log_filter=log_filter,
            need_format=False,
        )

    def _search_and_format(
        self,
        source: str = None,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = LOG_RECORDS_LIMIT,
        sort_by: str = "datetime",
        sort_order: str = SortOrder.DESC,
        log_filter: str = None,
        need_format: bool = True,
    ) -> List[Union[oci.loggingsearch.models.SearchResult, dict]]:
        """Returns the formatted consolidated log records.

        Parameters
        ----------
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        time_end : datetime.datetime, optional.
            Ending time for the query.
            Defaults to None.
        limit : int, optional.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
            Defaults to 100.
        sort_by : str, optional.
            The field for sorting the logs.
            Defaults to "datetime"
        sort_order : str, optional.
            The sort order for the log records. Can be "ASC" or "DESC".
            Defaults to "DESC".
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.
        need_format : bool, optional
            To decide whether to format the logs or not.
            Defaults to True.

        Returns
        -------
        list
            A list of oci.loggingsearch.models.SearchResult objects or log records.
            Each log record is a dictionary with the following keys: `annotation`, `id`, `time`,
            `message` and `datetime`.
        """
        if not time_start:
            time_start = datetime.datetime.utcnow() - datetime.timedelta(days=14)

        batch_logs = []
        for instance in self.logging_instance:
            batch_logs.extend(
                self._collect_logs(
                    instance,
                    source=source,
                    time_start=time_start,
                    time_end=time_end,
                    limit=limit,
                    sort_by=sort_by,
                    sort_order=sort_order,
                    log_filter=log_filter,
                    need_format=need_format,
                )
            )

        # _collect_logs returns a list of either dict or oci.loggingsearch.models.SearchResult
        # objects based on `need_format` parameter, so below there are two cases for log sorting.
        if need_format:
            batch_logs.sort(key=lambda x: x.get("datetime"))
        else:
            batch_logs.sort(key=lambda x: x.data.get("datetime"))
        if limit and len(batch_logs) > limit:
            batch_logs = batch_logs[:limit]
        return batch_logs

    def _collect_logs(
        self,
        logging_instance: OCILog,
        source: str = None,
        time_start: datetime.datetime = None,
        time_end: datetime.datetime = None,
        limit: int = LOG_RECORDS_LIMIT,
        sort_by: str = "datetime",
        sort_order: str = SortOrder.DESC,
        log_filter: str = None,
        need_format: bool = True,
    ) -> List[Union[oci.loggingsearch.models.SearchResult, dict]]:
        """Collects logs and formats log records.

        Parameters
        ----------
        logging_instance : OCILog
            OCILog instance.
        source : str, optional
            Expression or OCID to filter the "source" field of the OCI log record.
            Defaults to None.
        time_start : datetime.datetime, optional
            Starting time for the log query.
            Defaults to None.
        time_end : datetime.datetime, optional.
            Ending time for the query.
            Defaults to None.
        limit : int, optional.
            Maximum number of records to be returned.
            If limit is set to None, all logs from time_start to now will be returned.
            Defaults to 100.
        sort_by : str, optional.
            The field for sorting the logs.
            Defaults to "datetime"
        sort_order : str, optional.
            The sort order for the log records. Can be "ASC" or "DESC".
            Defaults to "DESC".
        log_filter : str, optional
            Expression for filtering the logs. This will be the WHERE clause of the query.
            Defaults to None.
        need_format : bool, optional
            To decide whether to format the logs or not.
            Defaults to True.

        Returns
        -------
        list
            A list of oci.loggingsearch.models.SearchResult objects or log records.
            Each log record is a dictionary with the following keys: `annotation`, `id`, `time`,
            `message` and `datetime`.
        """
        logs = logging_instance.search(
            source=source,
            time_start=time_start,
            time_end=time_end,
            limit=limit,
            sort_by=sort_by,
            sort_order=sort_order,
            log_filter=log_filter,
        )
        if logs and need_format:
            return [
                self._format_and_add_annotation(logging_instance.annotation, log.data)
                for log in logs
            ]

        return logs

    @staticmethod
    def _format_and_add_annotation(annotation: str, data) -> Dict:
        """Formats log and adds annotation.

        Parameters
        ----------
        annotation : str
            The logging annotation of OCILog instance.
        data : SearchResult.data
            The data from oci.loggingsearch.models.SearchResult

        Returns
        -------
        dict:
            Log record dictionary with the following keys: `annotation`, `id`, `time`,
            `message` and `datetime`.
        """
        log_content = data.get("logContent", {})
        return {
            "annotation": annotation,
            "id": log_content.get("id", ""),
            "message": log_content.get("data", {}).get("message", ""),
            "time": log_content.get("time", ""),
            "datetime": data.get("datetime", ""),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class Parser:
    @property
    def pos(self):
        raise NotImplementedError()

    @property
    def noun(self):
        raise NotImplementedError()

    @property
    def adjective(self):
        raise NotImplementedError()

    @property
    def adverb(self):
        raise NotImplementedError()

    @property
    def verb(self):
        raise NotImplementedError()

    @property
    def word(self):
        raise NotImplementedError()

    @property
    def sentence(self):
        raise NotImplementedError()

    @property
    def word_count(self):
        raise NotImplementedError()

    @property
    def bigram(self):
        raise NotImplementedError()

    @property
    def trigram(self):
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import absolute_import, print_function

import abc
import importlib
from collections import defaultdict
from numbers import Number
from typing import Union

import pandas as pd
from ads.common import utils, logger
from ads.common.data import ADSData
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.dataset import helper
from ads.dataset.dataset import ADSDataset
from ads.dataset.feature_engineering_transformer import FeatureEngineeringTransformer
from ads.dataset.feature_selection import FeatureImportance
from ads.dataset.helper import deprecate_default_value, deprecate_variable
from ads.dataset.label_encoder import DataFrameLabelEncoder
from ads.dataset.pipeline import TransformerPipeline
from ads.dataset.progress import DummyProgressBar
from ads.dataset.recommendation import Recommendation
from ads.dataset.recommendation_transformer import RecommendationTransformer
from ads.dataset.target import TargetVariable
from ads.type_discovery.typed_feature import DateTimeTypedFeature
from sklearn.model_selection import train_test_split
from pandas.io.formats.printing import pprint_thing
from sklearn.preprocessing import FunctionTransformer
from abc import ABCMeta


class ADSDatasetWithTarget(ADSDataset, metaclass=ABCMeta):
    """
    This class provides APIs for preparing dataset for modeling.
    """

    def __init__(
        self,
        df,
        sampled_df,
        target,
        target_type,
        shape,
        sample_max_rows=-1,
        type_discovery=True,
        types={},
        parent=None,
        name="",
        metadata=None,
        transformer_pipeline=None,
        description=None,
        progress=DummyProgressBar(),
        **kwargs,
    ):
        self.recommendation_transformer = None

        if parent is None:
            cols = sampled_df.columns.tolist()
            cols.insert(0, cols.pop(cols.index(target)))
            ADSDataset.__init__(
                self,
                df,
                sampled_df[[*cols]],
                shape,
                name=name,
                description=description,
                type_discovery=type_discovery,
                types=types,
                progress=progress,
                metadata=metadata,
                transformer_pipeline=transformer_pipeline,
                sample_max_rows=sample_max_rows,
            )
        else:
            self.__dict__ = parent.__dict__.copy()
            cols = self.sampled_df.columns.tolist()
            cols.insert(0, cols.pop(cols.index(target)))

            self.sampled_df = parent.sampled_df[[*cols]]

            # if parent has already been built, just reorder the columns to display the plot for target at beginning
            if parent.correlation is None:
                self.corr_futures = parent.corr_futures
            else:
                corr_cols = parent.sampled_df.select_dtypes(
                    exclude=["object"]
                ).columns.values.tolist()
                corr_cols.insert(0, corr_cols.pop(corr_cols.index(target)))
                self.correlation = parent.correlation.reindex(corr_cols)[[corr_cols]]
            self.feature_types = parent.feature_types
            self.feature_dist_html_dict = {}
            if len(parent.feature_dist_html_dict) > 0:
                parent_feature_dist_html_dict = parent.feature_dist_html_dict.copy()
                self.feature_dist_html_dict = {
                    target: parent_feature_dist_html_dict.pop(target)
                }
                self.feature_dist_html_dict.update(parent_feature_dist_html_dict)

        # drop all rows where target is nan
        target = target.strip().replace(" ", "_")

        #
        # as an optimization only dropna and regenerate sample when the target
        # has na values
        #

        if self.df[target].isna().sum():
            #
            # remove rows for which the target is null
            #
            self.df = self.df.dropna(subset=[target])

            #
            # we cannot simply drop null values from the sampled_df after a change
            # to the df - we must rebuild the sample from the new df
            #
            self.sampled_df = helper.generate_sample(
                self.df,
                sampled_df.shape[0],
                helper.DatasetDefaults.sampling_confidence_level,
                helper.DatasetDefaults.sampling_confidence_interval,
            )
            #
            # after regenerating the sample we need to move the target back to the head
            #
            cols = self.sampled_df.columns.tolist()
            cols.insert(0, cols.pop(cols.index(target)))
            self.sampled_df = self.sampled_df[[*cols]]

        self.target = TargetVariable(self, target, target_type)

        # remove target from type discovery conversion
        for step in self.transformer_pipeline.steps:
            if (
                step[0] == "type_discovery"
                and self.target.name in step[1].kw_args["dtypes"]
            ):
                step[1].kw_args["dtypes"].pop(self.target.name)

    def rename_columns(self, columns):
        """
        Returns a dataset with columns renamed.
        """
        if isinstance(columns, list):
            assert len(columns) == len(
                self.columns.values
            ), "columns length do not match the dataset"
            columns = dict(zip(self.columns.values, columns))
        assert isinstance(columns, dict)
        new_target = None
        if self.target.name in columns:
            new_target = columns[self.target.name]
        return self.rename(columns=columns, _new_target=new_target)

    def select_best_features(self, score_func=None, k=12):
        """
        Return new dataset containing only the top k features.

        Parameters
        ----------
        k: int, default 12
            The top 'k' features to select.
        score_func: function
            Scoring function to use to rank the features. This scoring function should take a 2d array X(features)
            and an array like y(target) and return a numeric score for each feature in the same order as X.

        Notes
        -----
        See also https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.f_regression.html
        and https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.f_classif.html

        Examples
        --------
        >>> ds = DatasetBrowser("sklearn").open("iris")
        >>> ds_small = ds.select_best_features(k=2)
        """
        tf = self._get_best_features_transformer(score_func=score_func, k=k)
        return self._build_new_dataset(
            tf[1].transform(self.df), tf[1].transform(self.sampled_df), transformers=tf
        )

    def auto_transform(
        self,
        correlation_threshold: float = 0.7,
        frac: float = 1.0,
        sample_size=1.0,
        correlation_methods: Union[str, list] = "pearson",
    ):
        """
        Return transformed dataset with several optimizations applied automatically.
        The optimizations include:

        - Dropping constant and primary key columns, which has no predictive quality,
        - Imputation, to fill in missing values in noisy data:

            - For continuous variables, fill with mean if less than 40% is missing, else drop,
            - For categorical variables, fill with most frequent if less than 40% is missing, else drop,

        - Dropping strongly co-correlated columns that tend to produce less generalizable models.

        Parameters
        ----------
        correlation_threshold: float, defaults to 0.7. It must be between 0 and 1, inclusive
            the correlation threshold where columns with correlation higher than the threshold will
            be considered as strongly co-correlated and recommended to be taken care of.
        frac: Is superseded by sample_size
        sample_size: float, defaults to 1.0. Float, Range -> (0, 1]
            What fraction of the data should be used in the calculation?
        correlation_methods: Union[list, str], defaults to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].

        Returns
        -------
        transformed_dataset : ADSDatasetWithTarget

        Examples
        --------
        >>> ds_clean = ds.auto_transform()
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            "<code>frac=None</code> is deprecated. Use <code>sample_size=1.0</code> instead.",
            FutureWarning,
        )

        if frac != 1.0:
            deprecate_frac = deprecate_variable(
                frac,
                sample_size,
                "<code>frac</code> is superseded by <code>sample_size</code>.",
                DeprecationWarning,
            )
            if sample_size == 1.0:
                sample_size = deprecate_frac

        if correlation_threshold > 1 or correlation_threshold < 0:
            raise AssertionError("correlation_threshold has to be between 0 and 1.")
        with utils.get_progress_bar(5) as progress:
            df, sampled_df, transformer_pipeline = self._transform(
                progress=progress,
                correlation_threshold=correlation_threshold,
                frac=sample_size,
                correlation_methods=correlation_methods,
            )
            return self._build_new_dataset(
                df,
                sampled_df=sampled_df,
                transformers=transformer_pipeline.steps,
                progress=progress,
            )

    def visualize_transforms(self):
        """
        Render a representation of the dataset's transform DAG.
        """

        helper.visualize_transformation(
            self.transformer_pipeline,
            text="- rows: {}\\l- columns: {}\\l".format(*self.shape),
        )

    def _suggested_code(self, action, recommendation_type, variable):
        if action == "Drop":
            return ".drop_columns([{}])".format('"' + variable + '"')
        if action == "Do nothing":
            return ""
        if "Drop " in action:
            return ".drop_columns([{}])".format('"' + action.split(" ")[1] + '"')
        if action == "Down-sample":
            return ".down_sample()"
        if action == "Up-sample":
            if importlib.util.find_spec("imblearn") is None:
                return ".up_sample(sampler='default') \\n `pip install imbalanced-learn` to use default up-sampler."
            else:
                return ".up_sample(sampler='default')"
        if recommendation_type == "positive_class" and action != "Do nothing":
            return ".set_positive_class({}, missing_value=False)".format(
                '"' + action + '"'
            )
        if recommendation_type == "imputation":
            fill_val = helper.get_fill_val(
                self.feature_types, variable, action, constant="constant"
            )

            fill_val = (
                fill_val if isinstance(fill_val, Number) else '"' + fill_val + '"'
            )
            return ".fillna({}{}: {}{})".format(
                "{", '"' + variable + '"', fill_val, "}"
            )
        else:
            return ""

    def suggest_recommendations(
        self,
        correlation_methods: Union[str, list] = "pearson",
        print_code: bool = True,
        correlation_threshold: float = 0.7,
        overwrite: bool = None,
        force_recompute: bool = False,
        frac: float = 1.0,
        sample_size: float = 1.0,
        **kwargs,
    ):
        """
        Returns a pandas dataframe with suggestions for dataset optimization. This includes:

        - Identifying constant and primary key columns, which has no predictive quality,
        - Imputation, to fill in missing values in noisy data:

            - For continuous variables, fill with mean if less than 40% is missing, else drop,
            - For categorical variables, fill with most frequent if less than 40% is missing, else drop,

        - Identifying strongly co-correlated columns that tend to produce less generalizable models,
        - Automatically balancing dataset for classification problems using up or down sampling.

        Parameters
        ----------
        correlation_methods: Union[list, str], default to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v']
        print_code: bool, Defaults to True
            Print Python code for the suggested actions.
        correlation_threshold: float. Defaults to 0.7. It must be between 0 and 1, inclusive
            the correlation threshold where columns with correlation higher than the threshold will
            be considered as strongly co-correated and recommended to be taken care of.
        frac: Is superseded by sample_size
        sample_size: float, defaults to 1.0. Float, Range -> (0, 1]
            What fraction of the data should be used in the calculation?
        overwrite:
            Is deprecated and replaced by force_recompute.
        force_recompute: bool, default to be False

            - If False, it calculates the correlation matrix if there is no cached correlation matrix. Otherwise,
              it returns the cached correlation matrix.
            - If True, it calculates the correlation matrix regardless whether there is cached result or not.

        Returns
        -------
        suggestion dataframe : pandas.DataFrame

        Examples
        --------
        >>> suggestion_df = ds.suggest_recommendations(correlation_threshold=0.7)
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            "<code>frac=None</code> is deprecated. Use <code>sample_size=1.0</code>.",
            FutureWarning,
        )

        if frac != 1.0:
            deprecate_frac = deprecate_variable(
                frac,
                sample_size,
                "<code>frac</code> is superseded by <code>sample_size</code>.",
                DeprecationWarning,
            )
            if sample_size == 1.0:
                sample_size = deprecate_frac

        force_recompute = deprecate_variable(
            overwrite,
            force_recompute,
            f"<code>overwrite=None</code> is deprecated. Use <code>force_recompute</code> instead.",
            DeprecationWarning,
        )

        recommended = self._get_recommendations_transformer(
            auto_transform=False,
            correlation_threshold=correlation_threshold,
            correlation_methods=correlation_methods,
            force_recompute=force_recompute,
            frac=sample_size,
            **kwargs,
        ).fit(self.sampled_df)

        if len(recommended.reco_dict_) == 0:
            logger.info("No recommendations.")
            return pd.DataFrame()

        column_names = [
            "Message",
            "Variables",
            "Action",
            "Selected Action",
            "Recommendation Type",
        ]

        df_dict = defaultdict(list)

        for recommendation_type, column_dict in recommended.reco_dict_.items():
            if recommendation_type == "constant_column":
                n_constant = len(column_dict)
                df_dict["Recommendation Type"].extend(
                    [recommendation_type] * n_constant
                )
                df_dict["Variables"].extend(column_dict)
                df_dict["Message"].extend(["Constant Column"] * n_constant)
                df_dict["Action"].extend(["Drop"] * n_constant)
                df_dict["Selected Action"].extend(["Drop"] * n_constant)
                continue

            for column, details_dict in column_dict.items():
                max_length = len(details_dict["Action"])
                for key, value in details_dict.items():
                    if isinstance(value, list):
                        df_dict[key].extend(value)
                    else:
                        df_dict[key].extend([value] * max_length)
                df_dict["Recommendation Type"].extend(
                    [recommendation_type] * max_length
                )
                df_dict["Variables"].extend([column] * max_length)

        suggestions_df = pd.DataFrame.from_dict(df_dict)[column_names]
        suggestions_df["Code"] = suggestions_df.apply(
            lambda x: self._suggested_code(
                x["Action"], x["Recommendation Type"], x["Variables"]
            ),
            axis=1,
        )
        suggestion_df = (
            suggestions_df.drop(columns=["Recommendation Type"])
            .rename(columns={"Selected Action": "Suggested"})
            .set_index(["Message", "Variables", "Suggested", "Action"])
            .fillna("")
        )
        if print_code:
            columns_to_impute = {}
            columns_to_drop = []
            consolidated_code = ""
            suggestion_df_ = suggestion_df.reset_index()
            suggested_code = suggestion_df_.loc[
                suggestion_df_.Suggested == suggestion_df_.Action
            ].Code.unique()
            for code in suggested_code:
                if ".drop_columns" in code:
                    columns_to_drop.append(code.split("[")[1].split("]")[0][1:-1])
                elif ".fillna" in code:
                    impute_pair = code.split("{")[1].split("}")[0]
                    columns_to_impute[impute_pair.split(":")[0].replace('"', "")] = (
                        float(impute_pair.split(":")[1].strip())
                        if impute_pair.split(":")[1].strip().replace(".", "").isdigit()
                        else impute_pair.split(":")[1].strip().replace('"', "")
                    )
                else:
                    consolidated_code += code
            consolidated_code = (
                "No more!" if len(consolidated_code) == 0 else consolidated_code
            )

            logger.info(f"Suggested columns to drop: {columns_to_drop}.")
            logger.info(f"Suggested columns to impute: {columns_to_impute}.")
            logger.info(f"Others: {consolidated_code}.")

        return suggestion_df

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def get_recommendations(
        self,
        correlation_methods: str = "pearson",
        correlation_threshold: float = 0.7,
        frac: float = 1.0,
        sample_size: float = 1.0,
        overwrite: bool = None,
        force_recompute: bool = False,
        display_format: str = "widget",
    ):
        """
        Generate recommendations for dataset optimization. This includes:

        - Identifying constant and primary key columns, which has no predictive quality,
        - Imputation, to fill in missing values in noisy data:

            - For continuous variables, fill with mean if less than 40% is missing, else drop,
            - For categorical variables, fill with most frequent if less than 40% is missing, else drop,

        - Identifying strongly co-correlated columns that tend to produce less generalizable models,
        - Automatically balancing dataset for classification problems using up or down sampling.

        Parameters
        ----------
        correlation_methods: Union[list, str], default to 'pearson'

            - 'pearson': Use Pearson's Correlation between continuous features,
            - 'cramers v': Use Cramer's V correlations between categorical features,
            - 'correlation ratio': Use Correlation Ratio Correlation between categorical and continuous features,
            - 'all': Is equivalent to ['pearson', 'cramers v', 'correlation ratio'].

            Or a list containing any combination of these methods, for example, ['pearson', 'cramers v'].
        correlation_threshold: float, defaults to 0.7. It must be between 0 and 1, inclusive
            The correlation threshold where columns with correlation higher than the threshold will
            be considered as strongly co-correlated and recommended to be taken care of.
        frac: Is superseded by sample_size
        sample_size: float, defaults to 1.0. Float, Range -> (0, 1]
            What fraction of the data should be used in the calculation?
        overwrite:
            Is deprecated and replaced by force_recompute.
        force_recompute: bool, default to be False

            - If False, it calculates the correlation matrix if there is no cached correlation matrix. Otherwise,
              it returns the cached correlation matrix.
            - If True, it calculates the correlation matrix regardless whether there is cached result or not.

        display_format: string, defaults to 'widget'.
            Should be either 'widget' or 'table'. If 'widget',
            a GUI style interface is popped out; if 'table', a table of suggestions is shown.
        """
        frac = deprecate_default_value(
            frac,
            None,
            1,
            "<code>frac=None</code> is superseded by <code>sample_size=1.0</code>.",
            FutureWarning,
        )

        if frac != 1.0:
            deprecate_frac = deprecate_variable(
                frac,
                sample_size,
                "<code>frac</code> is superseded by <code>sample_size</code>.",
                DeprecationWarning,
            )
            if sample_size == 1.0:
                sample_size = deprecate_frac

        force_recompute = deprecate_variable(
            overwrite,
            force_recompute,
            f"<code>overwrite=None</code> is deprecated. Use <code>force_recompute</code> instead.",
            DeprecationWarning,
        )

        if display_format == "widget":
            recommended = Recommendation(
                self,
                self._get_recommendations_transformer(
                    auto_transform=False,
                    correlation_threshold=correlation_threshold,
                    correlation_methods=correlation_methods,
                    frac=sample_size,
                    force_recompute=force_recompute,
                ).fit(self.sampled_df),
            )

            if len(recommended.reco_dict) == 0:
                logger.info("No recommendations.")

            return recommended

        elif display_format == "table":
            df_suggestion = self.suggest_recommendations(
                correlation_threshold=correlation_threshold,
                frac=sample_size,
                force_recompute=force_recompute,
            )

            from IPython.display import HTML, display

            display(
                HTML(
                    df_suggestion.to_html()
                    .replace(" `", "<code>")
                    .replace("` ", "</code>")
                    .replace("\\n", "<br>")
                )
            )

    def get_transformed_dataset(self):
        """
        Return the transformed dataset with the recommendations applied.

        This method should be called after applying the recommendations using the Recommendation#show_in_notebook() API.
        """
        if hasattr(self, "new_ds"):
            return self.new_ds
        logger.info(
            "Use `get_recommendations()` to view or update recommendation or `auto_tranform()` first."
        )
        logger.warning(
            "`get_transformed_dataset` is deprecated and will be removed in a future release."
        )
        return

    def type_of_target(self):
        """
        Return the target type for the dataset.

        Returns
        -------
        target_type: TypedFeature
            an object of TypedFeature

        Examples
        --------
        >>> ds = ds.set_target('target_class')
        >>> assert(ds.type_of_target() == 'categorical')
        """
        return self.target.type

    def train_test_split(self, test_size=0.1, random_state=utils.random_state):
        """
        Splits  dataset to train and test data.

        Parameters
        ----------
        test_size: Union[float, int], optional, default=0.1
        random_state: Union[int, RandomState], optional, default=None

                - If int, random_state is the seed used by the random number generator;
                - If RandomState instance, random_state is the random number generator;
                - If None, the random number generator is the RandomState instance used by np.random.

        Returns
        -------
        train_data, test_data: tuple
            tuple of ADSData instances

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> train, test = ds.train_test_split()
        """
        X_train, X_test, y_train, y_test = train_test_split(
            self.df.drop(self.target.name, axis=1),
            self.df[self.target.name],
            test_size=test_size,
            train_size=1 - test_size,
            random_state=random_state,
        )
        train = ADSData.build(
            X=X_train, y=y_train, name="Train Data", dataset_type=self.__class__
        )
        train.transformer_pipeline = self.transformer_pipeline
        test = ADSData.build(
            X=X_test, y=y_test, name="Test Data", dataset_type=self.__class__
        )
        return train, test

    def train_validation_test_split(
        self, test_size=0.1, validation_size=0.1, random_state=utils.random_state
    ):
        """
        Splits  dataset to train, validation and test data.

        Parameters
        ----------
        test_size: Union[float, int], optional, default=0.1
        validation_size: Union[float, int], optional, default=0.1
        random_state: Union[int, RandomState], optional, default=None

                - If int, random_state is the seed used by the random number generator;
                - If RandomState instance, random_state is the random number generator;
                - If None, the random number generator is the RandomState instance used by np.random.

        Returns
        -------
        train_data, validation_data, test_data: tuple
            tuple of ADSData instances

        Examples
        --------
        >>> ds = DatasetFactory.open("data.csv")
        >>> train, valid, test = ds.train_validation_test_split()
        """
        train, test = self.train_test_split(
            test_size=test_size, random_state=random_state
        )
        X_train, X_valid, y_train, y_valid = train_test_split(
            train.X, train.y, test_size=validation_size, random_state=random_state
        )
        train.X = X_train
        train.y = y_train
        valid = ADSData.build(
            X=X_valid, y=y_valid, name="Validation Data", dataset_type=self.__class__
        )
        return train, valid, test

    """
    Internal methods
    """

    def __repr__(self):
        rows, cols = self.shape
        return f"{self.__class__.__name__}(target: {self.target.name}) {rows:,} rows, {cols:,} columns"

    def _transform(
        self,
        progress=DummyProgressBar(),
        fix_imbalance=True,
        correlation_threshold=0.7,
        frac=None,
        correlation_methods="pearson",
    ):
        progress.update("Building the transformer pipeline")
        if self.recommendation_transformer is None:
            transformer_pipeline = TransformerPipeline(
                steps=[
                    (
                        "recommendations",
                        self._get_recommendations_transformer(
                            fix_imbalance=fix_imbalance,
                            correlation_threshold=correlation_threshold,
                            frac=frac,
                            correlation_methods=correlation_methods,
                        ),
                    ),
                    (
                        "feature_engineering",
                        FeatureEngineeringTransformer(
                            feature_metadata=self.feature_types
                        ),
                    ),
                ]
            )
        else:
            # recommendations are already generated using get_recommendations().show_in_notebook() API
            transformer_pipeline = TransformerPipeline(
                steps=[
                    (
                        "feature_engineering",
                        FeatureEngineeringTransformer(
                            feature_metadata=self.feature_types
                        ),
                    )
                ]
            )
            transformer_pipeline.steps = [
                ("recommendations", self.recommendation_transformer)
            ] + transformer_pipeline.steps
        sampled_df = self.sampled_df.copy()
        self.recommendation_transformer = None
        df = self.df.copy()
        for step in transformer_pipeline.steps:
            progress.update("Applying transformation for %s" % step[0])
            sampled_df = step[1].fit_transform(sampled_df)
            df = step[1].transform(df, progress=progress, fit_transform=True)
        return df, sampled_df, transformer_pipeline

    def _get_best_features(self, score_func=None, k=12):
        if isinstance(self.target.type, DateTimeTypedFeature):
            return FeatureImportance._get_feature_ranking(
                self.sampled_df.copy(),
                self.target.name,
                self.type_of_target(),
                score_func=score_func,
                k=k,
            )
        else:
            return FeatureImportance._get_feature_ranking(
                self.sampled_df.copy(),
                self.target.name,
                self.type_of_target(),
                score_func=score_func,
                k=k,
            )

    def _get_best_features_transformer(self, score_func=None, k=12):
        feature_set = self._get_best_features(k=k, score_func=score_func)[
            "features"
        ].tolist()

        def _select_features(df, feature_names, target):
            if target in df.columns:
                feature_names = feature_names + [target]
            return df[feature_names]

        return (
            "select_{0}_best_features".format(k),
            FunctionTransformer(
                func=_select_features,
                validate=False,
                kw_args={"feature_names": feature_set, "target": self.target.name},
            ).fit(self.sampled_df),
        )

    def _get_recommendations_transformer(
        self,
        fix_imbalance=True,
        auto_transform=True,
        correlation_threshold=0.7,
        **kwargs,
    ):
        force_recompute = kwargs.pop("force_recompute", False)
        frac = kwargs.pop("frac", 1)
        correlation_methods = kwargs.pop("correlation_methods", "pearson")
        return RecommendationTransformer(
            feature_metadata=self.feature_types,
            correlation=self.corr(
                force_recompute=force_recompute,
                frac=frac,
                correlation_methods=correlation_methods,
                **kwargs,
            ),
            target=self.target.name,
            target_type=self.target.type,
            is_balanced=self.target.is_balanced(),
            feature_ranking=self._get_best_features(k=len(self.sampled_df)),
            fix_imbalance=fix_imbalance,
            len=self.__len__(),
            auto_transform=auto_transform,
            correlation_threshold=correlation_threshold,
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import re

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import DocumentTypedFeature, AddressTypedFeature


class DocumentDetector(AbstractTypeDiscoveryDetector):

    _min_cjk_chars_for_document = 100
    _min_words = 10
    _min_html_tags = 5

    _html_pattern = re.compile("<.*?>")

    _unicode_ranges = [
        {"from": ord(u"\u3300"), "to": ord(u"\u33ff")},  # compatibility ideographs
        {"from": ord(u"\ufe30"), "to": ord(u"\ufe4f")},  # compatibility ideographs
        {"from": ord(u"\uf900"), "to": ord(u"\ufaff")},  # compatibility ideographs
        {
            "from": ord(u"\U0002F800"),
            "to": ord(u"\U0002fa1f"),
        },  # compatibility ideographs
        {"from": ord(u"\u3040"), "to": ord(u"\u309f")},  # Japanese Hiragana
        {"from": ord(u"\u30a0"), "to": ord(u"\u30ff")},  # Japanese Katakana
        {"from": ord(u"\u2e80"), "to": ord(u"\u2eff")},  # cjk radicals supplement
        {"from": ord(u"\u4e00"), "to": ord(u"\u9fff")},
        {"from": ord(u"\u3400"), "to": ord(u"\u4dbf")},
        {"from": ord(u"\U00020000"), "to": ord(u"\U0002a6df")},
        {"from": ord(u"\U0002a700"), "to": ord(u"\U0002b73f")},
        {"from": ord(u"\U0002b740"), "to": ord(u"\U0002b81f")},
        {
            "from": ord(u"\U0002b820"),
            "to": ord(u"\U0002ceaf"),
        },  # included as of Unicode 8.0
    ]

    def _is_cjk_char(self, char):
        return any(
            [
                range["from"] <= ord(char) <= range["to"]
                for range in DocumentDetector._unicode_ranges
            ]
        )

    def cjk_string(self, document):
        cjk_char_count = sum([1 if self._is_cjk_char(c) else 0 for c in document])
        return cjk_char_count / len(document) >= 0.2

    def html_document(self, document):
        return (
            len(list(re.finditer(DocumentDetector._html_pattern, document)))
            > DocumentDetector._min_html_tags
        )

    def discover(self, name, series):
        #
        # very basic detection of a document. If the document is CJK then we use only the document length
        # otherwise we split on whitespace and confirm that there are word-like strings
        #
        if series.dtype == "object":
            null_series = series.loc[~series.isnull()]
            first_non_null_document = null_series.iloc[0]

            if isinstance(first_non_null_document, str):

                is_html = self.html_document(first_non_null_document)

                if self.cjk_string(first_non_null_document):
                    if (
                        len(first_non_null_document)
                        >= DocumentDetector._min_cjk_chars_for_document
                    ):
                        tf = DocumentTypedFeature.build(
                            name, series, is_cjk=True, is_html=is_html
                        )
                        logger.debug(
                            "type discovery on CJK column [{}]/[{}] found to be a document".format(
                                name, series.dtype
                            )
                        )
                        return tf
                else:
                    # find rows with above average length
                    above_avg_series = null_series.loc[
                        null_series.str.len() >= null_series.str.len().mean()
                    ]

                    # take a sample. max 500 docs
                    above_avg_series_sample = above_avg_series.sample(
                        n=min(500, len(above_avg_series))
                    )

                    # if all of the samples have more than min_words tokens..
                    mean_number_of_words = (
                        above_avg_series_sample.str.split().str.len().mean()
                    )
                    if mean_number_of_words > DocumentDetector._min_words:
                        if (
                            mean_number_of_words < 15
                            and above_avg_series_sample.str.count(",").mean()
                            / mean_number_of_words
                            > 0.1
                        ):
                            # many commas probably means address type
                            logger.debug(
                                "type discovery on column [{}]/[{}] looks like an address type".format(
                                    name, series.dtype
                                )
                            )
                            return AddressTypedFeature.build(name, series)
                        else:
                            logger.debug(
                                "type discovery on non-CJK column [{}]/[{}] found to be a document".format(
                                    name, series.dtype
                                )
                            )

                            # previous check of first document for HTML is now refined using longer documents
                            is_html = all(
                                [
                                    self.html_document(doc)
                                    for doc in above_avg_series_sample
                                ]
                            )

                            return DocumentTypedFeature.build(
                                name, series, is_cjk=False, is_html=is_html
                            )

        return False


if __name__ == "__main__":
    dd = DocumentDetector()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import base64
import json
import os
from typing import Dict
from typing import Tuple

import yaml
import glob

from ads.common.auth import create_signer
from ads.opctl import logger
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.utils import NotSupportedError, convert_notebook
from ads.opctl.constants import (
    ML_JOB_GPU_IMAGE,
    ML_JOB_IMAGE,
    BACKEND_NAME,
)
from ads.opctl.utils import (
    list_ads_operators,
    parse_conda_uri,
    get_region_key,
    get_namespace,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class ConfigResolver(ConfigProcessor):
    """Infer and fill in fields necessary for running an operator based on user inputs.
    For instance, if user pass in an ADS operator name, we will fill in image name, command,
    volumes to mount for local run, source folder path, etc.

    Please check functions with name `_resolve_<thing>`.

    List of things being resolved (and brief description):
    - operator name (could be passed in directly or in YAML)
    - conda (type is determined on whether a slug or uri is passed in. slug -> service, uri -> published. also do some additional checks.)
    - source folder (mainly for ADS operator)
    - command (docker command)
    - image name (mainly for ADS operator)
    - env vars
    - mount volumes (for local runs)

    There are additional comments in some of the functions below.
    """

    def __init__(self, config: Dict = None) -> None:
        super().__init__(config)
        self.ads_operators = list_ads_operators()

    def process(self):
        # this function should be run after merging configs
        # conda pack scenarios --
        # - user runs their own scripts and commands -> source_folder + entrypoint + (command)
        # - user runs ADS operator -> name/YAML

        # docker image scenarios --
        # - user runs their own docker image -> image (to run) + (entrypoint) + (command)
        # - user runs ADS operator -> name/YAML

        if self.config["execution"].get("job_id"):
            return self

        if self.config.get("kind") == "pipeline":
            # For pipelines, properties like the conda slug and source folder don't apply to the entire pipeline,
            # so we can skip all this validation. Each individual step will have its own config with the required
            # values set.
            return self

        self._resolve_operator_name()

        if not (
            self._is_ads_operator()
            or self.config["execution"].get("source_folder")
            or self.config["execution"].get("image")
        ):
            raise ValueError(
                "Either an ADS operator or a source folder or a docker image needs to be given."
            )

        self._resolve_conda()
        if not (
            self.config["execution"].get("conda_slug")
            or self.config["execution"].get("image")
            or self.config["execution"]["backend"] == BACKEND_NAME.DATAFLOW.value
        ):
            raise ValueError(
                "If not running an operator, conda pack info or image name needs to be given."
            )
        if self.config["execution"].get("conda_slug") and self.config["execution"].get(
            "image"
        ):
            raise ValueError(
                "Both conda pack info and image name are given. It is ambiguous which one to use."
            )

        self._resolve_source_folder_path()
        self._resolve_entry_script()
        self._resolve_command()
        self._resolve_image_name()
        self._resolve_env_vars()
        self._resolve_mounted_volumes()
        self._resolve_job_name()

        logger.debug(f"Config: {self.config}")

        return self

    def _is_ads_operator(self) -> bool:
        return self.config["execution"].get("operator_name") in self.ads_operators

    def _resolve_operator_name(self) -> None:
        self.config["execution"]["operator_name"] = self.config["execution"].get(
            "operator_name"
        ) or self.config.get("name")

    @runtime_dependency(module="inflection", install_from=OptionalDependency.OPCTL)
    def _resolve_source_folder_path(self) -> None:
        # this should be run after resolve_operator_name()
        # resolve ADS operator source folder path
        exec_config = self.config["execution"]
        if exec_config.get("source_folder"):
            self.config["execution"]["source_folder"] = os.path.abspath(
                os.path.normpath(exec_config["source_folder"])
            )
            if not os.path.exists(self.config["execution"]["source_folder"]):
                raise FileNotFoundError(
                    f"{self.config['execution']['source_folder']} is not found."
                )
        else:
            if self._is_ads_operator():
                curr_dir = os.path.dirname(os.path.abspath(__file__))
                self.config["execution"]["source_folder"] = os.path.normpath(
                    os.path.join(
                        curr_dir,
                        "..",
                        "operators",
                        inflection.underscore(
                            self.config["execution"]["operator_name"]
                        ),
                    )
                )
            else:
                self.config["execution"]["source_folder"] = None

    def _resolve_entry_script(self) -> None:
        # this should be run after _resolve_source_folder_path
        if not self._is_ads_operator():
            if os.path.splitext(self.config["execution"]["entrypoint"])[1] == ".ipynb":
                input_path = os.path.join(
                    self.config["execution"]["source_folder"],
                    self.config["execution"]["entrypoint"],
                )
                exclude_tags = self.config["execution"].get("exclude_tag")
                self.config["execution"]["entrypoint"] = os.path.basename(
                    convert_notebook(
                        input_path,
                        {},
                        exclude_tags,
                        overwrite=self.config["execution"].get("overwrite", False),
                    )
                )

    def _resolve_command(self) -> None:
        # this should be run after resolve_operator_name and resolve_conda
        # resolve command for operators
        exec_config = self.config["execution"]
        if self._is_ads_operator():
            logger.info("Building command for operator runs...")
            operator_name = exec_config.get("operator_name")
            command = f"-n {operator_name} -c {exec_config['oci_config']} -p {exec_config['oci_profile']}"
            if "spec" in self.config:
                encoded_spec = base64.b64encode(
                    json.dumps(self.config["spec"]).encode()
                ).decode()
                command += f" -s {encoded_spec}"
            self.config["execution"]["command"] = command
        else:
            # cmd_args is for running scripts, either locally, with jobs or dataflow; command is for docker command
            self.config["execution"]["command"] = exec_config.get(
                "cmd_args"
            ) or exec_config.get("command")

    def _resolve_conda(self) -> None:
        # this should be run after resolve_operator_name
        exec_config = self.config["execution"]
        # two scenarios that indicate conda pack should be used:
        # - conda slug is found
        # - use-conda is turned on explicitly. in this case for ads operators, look up index.yaml to fill in conda info
        slug = None
        conda_type = None
        if exec_config.get("conda_uri"):
            slug = parse_conda_uri(exec_config["conda_uri"])[-1]
            conda_type = "published"
        elif exec_config.get("conda_slug"):
            slug = exec_config["conda_slug"]
            conda_uri, conda_type = self._get_conda_uri_and_type(slug)
            self.config["execution"]["conda_uri"] = conda_uri
            if conda_type and conda_type == "data_science":
                conda_type = "service"
            conda_type = conda_type or "service"
        elif exec_config.get("use_conda", False):
            operator_name = exec_config.get("operator_name")
            if self._is_ads_operator():
                if self.ads_operators[operator_name].get("conda_slug"):
                    slug = self.ads_operators[operator_name]["conda_slug"]
                    conda_type = "service"
                else:
                    raise NotSupportedError(
                        f"Conda pack is not supported for ADS operator {operator_name}."
                    )
        else:
            logger.info("Conda pack slug info is not found.")
        self.config["execution"]["conda_slug"] = slug
        self.config["execution"]["conda_type"] = conda_type

    def _get_conda_uri_and_type(self, slug: str) -> Tuple[str, str]:
        conda_pack_path = os.path.join(
            os.path.expanduser(self.config["execution"]["conda_pack_folder"]), slug
        )
        if os.path.exists(conda_pack_path):
            manifest_path = glob.glob(os.path.join(conda_pack_path, "*_manifest.yaml"))[
                0
            ]
            with open(manifest_path) as f:
                manifest = yaml.safe_load(f.read())["manifest"]
            return manifest.get("pack_uri"), manifest.get("type")
        else:
            return None, None

    def _resolve_image_name(self) -> None:
        # this should be run after resolve_conda() and resolve_command()
        exec_config = self.config["execution"]
        if exec_config.get("conda_slug", None):
            self.config["execution"]["image"] = (
                ML_JOB_GPU_IMAGE if exec_config.get("gpu", False) else ML_JOB_IMAGE
            )
        elif self._is_ads_operator():
            image = self.ads_operators[exec_config["operator_name"]].get("image")
            oci_auth = create_signer(
                self.config["execution"].get("auth"),
                self.config["execution"].get("oci_config"),
                self.config["execution"].get("oci_profile"),
            )
            region_key = get_region_key(oci_auth).lower()
            if exec_config.get("namespace"):
                namespace = exec_config["namespace"]
            else:
                namespace = get_namespace(oci_auth)
            if self.config.get("infrastructure", {}).get("docker_registry"):
                self.config["execution"]["image"] = os.path.join(
                    self.config["infrastructure"]["docker_registry"], image
                )
            else:
                self.config["execution"][
                    "image"
                ] = f"{region_key}.ocir.io/{namespace}/{image}"

    def _resolve_env_vars(self) -> None:
        env_vars = self.config["execution"].get("env_vars", {})
        if self.config["execution"].get("env_var"):
            for ev in self.config["execution"]["env_var"]:
                if "=" in ev:
                    k, v = ev.split("=")
                    env_vars[k.strip()] = v.strip()
                elif ev in os.environ:
                    env_vars[ev] = os.environ[ev]
                else:
                    raise ValueError(
                        "Please provide environment variable as VAR=value or "
                        "give an environment variable that is set locally."
                    )
        self.config["execution"]["env_vars"] = env_vars

    def _resolve_mounted_volumes(self) -> None:
        volumes = self.config["execution"].get("volumes", {})
        if self.config["execution"].get("volume"):
            for v in self.config["execution"]["volume"]:
                host, remote = v.split(":")
                if not os.path.exists(host.strip()):
                    raise FileNotFoundError(f"{host} does not exist")
                volumes[os.path.abspath(host.strip())] = {"bind": remote.strip()}
        self.config["execution"]["volumes"] = volumes

    def _resolve_job_name(self) -> None:
        # this should be run after resolve_operator_name
        if not self.config["execution"].get("job_name"):
            if self.config["execution"].get("operator_name"):
                self.config["execution"]["job_name"] = self.config["execution"][
                    "operator_name"
                ]
